/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package beagleutil;

import java.util.Arrays;
import blbutil.*;

/**
 * Class <code>GprobsData</code> represented a line of data
 * from a file in BEAGLE .gprobs file format.
 * 
 * @author Brian L Browning
 */
public class GprobsData {

    private final Samples samples;
    private final String markerId;
    private final int nAlleles;
    private final int nGenotypes;
    private final String[] alleles;
    private final double[] data;
    private final double[] normData;

    /**
     * Constructs a <code>GprobsData</code> object from the specified 
     * data line of a BEAGLE .gprobs file.
     *
     * @param line a non-header line of data from a file in BEAGLE .gprobs
     * format.
     * @param samples the filtered and unfiltered sample identifiers for the
     * data.
     * @param isMultiAllelic <code>true</code> if the specified
     * <code>line</code> contains a field specifying the number of alleles
     * for the marker and <code>false</code> otherwise.
     * If <code>isMultiAllelic==true</code> the second field
     * of the line contains the number of alleles, and the third field
     * begins the enumeration of the marker alleles.
     * If <code>isMultiAllelic==false</code> the marker is assumed to be
     * a diallelic marker, and the second and third fields give the marker
     * alleles.
     *
     * @throws NullPointerException if <code>line==null || samples==null</code>.
     * @throws IllegalArgumentException 
     * if the number of alleles is less than 1, or if any
     * genotype probabilty is negative, or if if the
     * number of fields on the line is inconsistent with the number of
     * specified alleles and the number of specified unfiltered samples.
     * @throws NumberFormatException if any field that is expected to be an
     * integer or a double is not parsable.
     */
    public GprobsData(String line, Samples samples, boolean isMultiAllelic) {
        String[] fields = StringUtil.getFields(line);
        this.nAlleles = checkArguments(fields, samples, isMultiAllelic);
        this.samples = samples;
        this.markerId = fields[0];
        this.nGenotypes = (this.nAlleles * (this.nAlleles + 1))/2;
        this.alleles = new String[this.nAlleles];
        int index = isMultiAllelic ? 2 : 1;
        for (int j=0; j<this.alleles.length; ++j) {
            this.alleles[j] = fields[index++];
        }
        this.data = new double[this.samples.filteredSize()*this.nGenotypes];
        this.normData = new double[this.data.length];
        double[] values = parseAndFilterData(fields, index);
        fillArrays(values);
    }

    /**
     * Constructs a <code>GprobsData</code> object obtained by replacing
     * the per-sample data and normalized data in a specified
     * <code>GprobsData</code> instance.
     *
     * @param gprobsData a <code>GprobsData</code> object.
     * @param updatedData an array of non-negative numbers of length
     * <code>gprobsData.nGenotypes()*gprobsData.nSamples()</code>.
     *
     * @throws NullPointerException if 
     * <code>gprobsData==null || updatedData==null</code>.
     * @throws IllegalArgumentException
     * if <code>gprobsData.data().length != updatedData.length</code> or if
     * any element of the specified <code>updatedData</code> array is negative.
     */
    public GprobsData(GprobsData gprobsData, double[] updatedData) {
        if (updatedData.length != gprobsData.data.length) {
            int size = gprobsData.nGenotypes()*gprobsData.nSamples();
            String s = "inconsistent length for marker " + gprobsData.markerId()
                    + " expected=" + size + " found=" + updatedData.length;
            throw new IllegalArgumentException(s);
        }
        this.nAlleles = gprobsData.nAlleles;
        this.samples = gprobsData.samples;
        this.markerId = gprobsData.markerId;
        this.nGenotypes = gprobsData.nGenotypes;
        this.alleles = gprobsData.alleles;
        this.data = new double[gprobsData.data.length];
        this.normData = new double[gprobsData.data.length];
        fillArrays(updatedData);
    }

    private GprobsData(GprobsData gprobsData, boolean swapAlleles) {
        this.nAlleles = gprobsData.nAlleles;
        this.samples = gprobsData.samples;
        this.markerId = gprobsData.markerId;
        this.nGenotypes = gprobsData.nGenotypes;
        this.alleles = gprobsData.alleles;
        this.data = gprobsData.data.clone();
        this.normData = gprobsData.normData.clone();
        if (swapAlleles) {
            if (this.alleles.length != 2) {
                throw new IllegalArgumentException("expected 2 alleles: "
                        + gprobsData);
            }
            String s = this.alleles[0];
            this.alleles[0] = this.alleles[1];
            this.alleles[1] = s;
            for (int j=0; j<data.length; j+=3) {
                double tmp = data[j];
                data[j] = data[j + 2];
                data[j + 2] = tmp;

                tmp = normData[j];
                normData[j] = normData[j+2];
                normData[j+2] = tmp;
            }
        }
    }

    public GprobsData swapAlleles() {
        boolean swapAlleles = true;
        return new GprobsData(this, swapAlleles);
    }

    /*
     * Returns the number of alleles for the marker.
     */
    private int checkArguments(String[] fields, Samples samples,
            boolean isMultiAllelic) {
        assert fields != null;
        int alleleCnt = isMultiAllelic ? Integer.parseInt(fields[1]) : 2;
        if (alleleCnt < 1) {
            String s = "nAlleles < 1: " + alleleCnt + ": " +
                    Arrays.toString(fields);
            throw new IllegalArgumentException(s);
        }
        int genotypeCnt = (alleleCnt * (alleleCnt + 1))/2;
        int expectedFields = 1 + alleleCnt
                + genotypeCnt*samples.unfilteredSize();
        if (isMultiAllelic) {
            ++expectedFields;
        }
        if (fields.length != expectedFields) {
            String s = "Error for marker " + fields[0] + ": expected "
                    + expectedFields + " fields, found " + fields.length +
                    " fields (nSamples=" + samples.unfilteredSize() 
                    + " nAlleles=" + nAlleles + "): " + Arrays.toString(fields);
            throw new IllegalArgumentException(s);
        }
        return alleleCnt;
    }

    private double[] parseAndFilterData(String[] fields, int startIndex) {
        double[] values = new double[samples.filteredSize() * this.nGenotypes];
        int sampleIndex = 0;
        int valuesIndex = 0;
        for (int j=startIndex; j<fields.length; j+=this.nGenotypes) {
            if (this.samples.filter(sampleIndex++)==false) {
                for (int k=0; k<this.nGenotypes; ++k) {
                    values[valuesIndex++] = Double.parseDouble(fields[j+k]);
                }
            }
        }
        assert valuesIndex==values.length;
        return values;
    }

    private void fillArrays(double[] values) {
        double missing = 1.0/this.nGenotypes;
        for (int j=0; j<values.length; j+=this.nGenotypes) {
            double sum = 0.0;
            for (int k=0; k<this.nGenotypes; ++k) {
                if (values[j+k] < 0) {
                    String s = "Negative probability for marker "
                            + this.markerId + ": " + values[j+k];
                    throw new IllegalArgumentException(s);
                }
                this.data[j+k] = values[j+k];
                sum += values[j+k];
            }
            if (sum == 0.0) {
                for (int k=0; k<this.nGenotypes; ++k) {
                    this.normData[j+k] = missing;
                }
            }
            else {
                for (int k=0; k<this.nGenotypes; ++k) {
                    this.normData[j+k] = this.data[j+k]/sum;
                }
            }
        }
    }

    /**
     * Returns the marker identifier.
     * @return the marker identifier.
     */
    public String markerId() {
        return markerId;
    }

    /**
     * Returns the number of samples.
     * @return the number of samples.
     */
    public int nSamples() {
        return samples.filteredSize();
    }

    /**
     * Returns the specified sample identifier.
     * @param k an index of an sample.
     * @return the specified sample identifier.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.Samples()</code>.
     */
    public String sampleIds(int k) {
        return samples.filteredIds(k);
    }

    /**
     * Returns an array of sample identifiers such that
     * <code>this.samples()[k].equals(this.sample(k))</code>
     * for <code> 0 &le; k < this.filteredSize()</code>.
     *
     * @return an array of sample identifiers.
     */
    public String[] sampleIds() {
        return samples.filteredIds();
    }


   /**
     * Returns the number of allele identifiers for the marker.
     * @return the number of allele identifiers for the marker.
     */
    public int nAlleles() {
        return nAlleles;
    }

    /**
     * Returns the number of genotype identifiers for the marker
     * @return the number of genotype identifiers for the marker.
     */
    public int nGenotypes() {
        return nGenotypes;
    }

    /**
     * Returns an array of allele identifiers.
     * The <code>this.alleles()[k].equals(this.alleles(k))</code>
     * for <code> 0 &le; k < this.nAlleles()</code>.
     *
     * @return an array of allele identifiers.
     */
    public String[] alleles() {
        return alleles.clone();
    }
    
    /**
     * Returns the specified alleles identifier.
     * @param k an index of an alleles.
     * @return the specified alleles identifier.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nAlleles()</code>.
     */
    public String alleles(int k) {
        return alleles[k];
    }

    /**
     * Returns the array of probability data for the marker.
     * @return the array of probability data for the marker.
     */
    public double[] data() {
        return data.clone();
    }

    /**
     * Returns the normalized array of probability data for the marker.
     * The sum of the probabilities is 1.0 for each sample for the normalized
     * data.
     * @return the normalized array of probability data for the marker.
     */
    public double[] normData() {
        return normData.clone();
    }

    /**
     * Returns <code>data()[j]</code>.
     * @param j an array index.
     * @return <code>data()[j]</code>.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nSamples()*this.nGenotypes()</code>.
     */
    public double data(int j) {
        return data[j];
    }

    /**
     * Returns <code>normalizedData()[j]</code>.
     * @param j an array index.
     * @return <code>normalizedData()[j]</code>.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nSamples()*this.nGenotypes()</code>.
     */
    public double normData(int j) {
        return normData[j];
    }

    @Override
    public String toString() {
        String space = " ";
        StringBuilder sb = new StringBuilder();
        sb.append(markerId);
        sb.append(space);
        sb.append(nAlleles);
        for (int j=0; j<nAlleles; ++j) {
            sb.append(space);
            sb.append(alleles(j));
        }
        sb.append(space);
        sb.append(nGenotypes);
        sb.append(space);
        sb.append(samples.filteredSize());
        for (int j=0, n=samples.filteredSize(); j<n; ++j) {
            sb.append(space);
            sb.append(samples.filteredIds(j));
            sb.append("=");
            sb.append(Arrays.toString(
                    Arrays.copyOfRange(data, nGenotypes*j, nGenotypes*(j+1))));
        }
        return sb.toString();
    }
}
