/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package beagleutil;

import java.io.File;
import java.util.*;
import blbutil.*;

/**
 * Class <code>GprobsIterator</code> iterates through data in a file
 * in BEAGLE .gprobs format and returns a <code>GprobsData</code> instance
 * for each marker in the .gprobs file.
 *
 * @author Brian L Browning
 */
public class GprobsIterator implements FileIterator<GprobsData> {

    private final Samples sampleIds;
    private final boolean isMultiAllelic;
    private final FileIterator<String> it;
    private final ImmutableSet<String> markerExclusions;

    private GprobsData next;

    /**
     * Constructs a <code>GprobsIterator</code> instance.
     *
     * @param gprobsFile a file in BEAGLE .gprobs format.  If
     * <code>isMultiAllelic==true</code> then the first (header) line of the
     * specified .gprobs file consists of a white-space delimited list
     * of ordered sample identifiers.  If <code>isMultiAllelic==false</code>
     * then the first (header) line begins with "marker alleleA alleleB "
     * and continues with the sample identifiers corresponding to each
     * column of input data (i.e. each sample identifier is repeated three
     * times).
     * @param sampleExclusions sample identifiers to be excluded.
     * @param markerExclusions marker identifiers to be excluded.
     * @param isMultiAllelic returns <code>true</code> if the input data
     * files contain a field specifying the number of alleles and
     * <code>false</code> otherwise.  If <code>isMultiAllelic==false</code>
     * the marker is assumed to be diallelic.
     *
     * @throws NullPointerException if
     * <code>gprobsFiles==null || sampleExclusions==null
     * || markerExclusions==null</code>.
     *
     * @throws IllegalArgumentException if the specified .gprobs file is not
     * in BEAGLE .gprobs file format.
     *
     */
    public GprobsIterator(File gprobsFile, 
            ImmutableSet<String> sampleExclusions,
            ImmutableSet<String> markerExclusions, boolean isMultiAllelic) {
        this(InputStreamIterator.fromGzipFile(gprobsFile), sampleExclusions,
                markerExclusions, isMultiAllelic);
    }

    /**
     * Constructs a <code>GprobsIterator</code> instance.
     *
     * @param it a <code>FileIterator<String></code> whose
     * <code>next()</code> method will return lines of a file in
     * BEAGLE .gprobs format.  If
     * <code>isMultiAllelic==true</code> then the first (header) line of the
     * specified .gprobs file consists of a white-space delimited list
     * of ordered sample identifiers.  If <code>isMultiAllelic==false</code>
     * then the first (header) line begins with "marker alleleA alleleB "
     * and continues with the sample identifiers corresponding to each
     * column of input data (i.e. each sample identifier is repeated three
     * times).
     * @param sampleExclusions sample identifiers to be excluded.
     * @param markerExclusions marker identifiers to be excluded.
     * @param isMultiAllelic returns <code>true</code> if the input data
     * files contain a field specifying the number of alleles and
     * <code>false</code> otherwise.  If <code>isMultiAllelic==false</code>
     * the marker is assumed to be diallelic.
     *
     * @throws NullPointerException if
     * <code>gprobsFiles==null || sampleExclusions==null
     * || markerExclusions==null</code>.
     *
     * @throws IllegalArgumentException if the specified .gprobs file is not
     * in BEAGLE .gprobs file format.
     *
     */
    public GprobsIterator(FileIterator<String> it,
            ImmutableSet<String> sampleExclusions,
            ImmutableSet<String> markerExclusions, boolean isMultiAllelic) {
        if (markerExclusions==null) {
            throw new NullPointerException("markerExclusions==null");
        }
        if (sampleExclusions==null) {
            throw new NullPointerException("sampleExclusions==null");
        }
        this.isMultiAllelic = isMultiAllelic;
        this.markerExclusions = markerExclusions;
        this.it = it;
        String headerLine = readFirstNonBlankLine(this.it);
        String[] ids = ids(headerLine, isMultiAllelic);
        this.sampleIds = new Samples(ids, sampleExclusions);
        this.next = readData();
    }


    private String readFirstNonBlankLine(FileIterator<String> it) {
        String line = null;
        while (it.hasNext() && line==null) {
            String candidate = it.next().trim();
            if (candidate.length() > 0) {
                line = candidate;
            }
        }
        if (line==null) {
            String s = " missing data in .gprobs file";
            throw new IllegalArgumentException(s);
        }
        return line;
    }

    private String[] ids(String headerLine, boolean isMultiAllelic) {
        String[] headerFields = StringUtil.getFields(headerLine);
        if (isMultiAllelic) {
            return headerFields;
        }
        else {
            int idIndex = 0;
            int startIndex = 3;
            if ((headerFields.length % 3) != 0) {
                String s = "number of fields in header line of "
                        + ".gprobs file is not divisible by 3: "
                        + headerFields;
                throw new IllegalArgumentException(s);
            }
            String[] ids = new String[(headerFields.length - 3) / 3];
            for (int j=startIndex; j<headerFields.length; j+=3) {
                for (int k=1; k<3; ++k) {
                    if (!headerFields[j].equals(headerFields[j + k])) {
                        String s = "The three columns of data for a sample " +
                                "do not have the same sample identifier " +
                                "in the first line of the .gprobs file: " +
                                headerFields[j] + " != " + headerFields[j+k];
                        throw new IllegalArgumentException(s);
                    }
                }
                ids[idIndex++] = headerFields[j];
            }
            assert idIndex == ids.length;
            return ids;
        }
    }

    public void close() {
        it.close();
        next = null;
    }

    public boolean hasNext() {
        return (next != null);
    }

    public GprobsData next() {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        GprobsData current = next;
        next = readData();
        return current;
    }

    public void remove() {
        String s = "remove() is not supported by GprobsIterator";
        throw new UnsupportedOperationException(s);
    }

    private GprobsData readData() {
        GprobsData data = null;
        while (it.hasNext() && data==null) {
            String candidate = it.next();
            String[] fields = StringUtil.getFields(candidate, 2);
            if (fields.length > 0 && !markerExclusions.contains(fields[0])) {
                data = new GprobsData(candidate, sampleIds, isMultiAllelic);
            }
        }
        return data;
    }

    /**
     * Returns a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to change.
     * @return a string representation of <code>this</code>.
     */
    @Override
    public String toString() {
        String space = " ";
        StringBuilder sb = new StringBuilder();
        sb.append("isMultiallelic=");
        sb.append(isMultiAllelic);
        sb.append(space);
        sb.append(" next=");
        sb.append(next);
        return sb.toString();
    }
}
