/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */package beagleutil;

import java.io.*;
import java.text.DecimalFormat;

/**
 * Class <code>GprobsPrinter</code> is a class for printing
 * <code>GprobsData</code> instances to a filename.  Output is written
 * in BEAGLE .gprobs format.  Distinct <code>GprobsPrinter</code>
 * instances must be used for distinct output files because the
 * <code>GprobsPrinter</code> automatically prints a header line
 * before printing the first data line.
 * 
 * @author Brian L Browning
 */
public final class GprobsPrinter {

    private static final String space = " ";
    private final DecimalFormat df;
    private final boolean isMultiAllelic;
    private boolean firstLine = true;

    /**
     * Constructs a <code>GprobsPrinter</code> instance.
     * 
     * @param df a <code>DecimalFormat</code> instance for formatting
     * numerical fields.
     * @param isMultiAllelic <code>true</code> if the output should be
     * written in multi-allelic format and <code>false</code> if the
     * output should be written in diallelic format.
     * @throws IllegalArgumentException if <code>nDecimalPlaces < 0</code>.
     */
    public GprobsPrinter(DecimalFormat df, boolean isMultiAllelic) {
        if (df==null) {
            throw new NullPointerException("df==null");
        }
        this.df = new DecimalFormat(df.toPattern());
        this.isMultiAllelic = isMultiAllelic;
    }

    /**
     * Prints a space delimited line in BEAGLE .gprobs format corresponding
     * to the data in the specified <code>GprobsData</code> instance.
     *
     * @param data a <code>GprobsData</code> instance to be printed.
     * @param out PrintWriter to receive output line.
     *
     * @throws IllegalStateException if the method is invoked after
     * <code>this.close()</code> method is invoked.
     */
    public void print(GprobsData data, PrintWriter out) {
        if (out==null) {
            throw new IllegalStateException("Error: Printer has been closed");
        }
        if (firstLine) {
            out.println(getOutputHeaderLine(data));
            firstLine = false;
        }
        out.print(data.markerId());
        if (isMultiAllelic) {
            out.print(space);
            out.print(data.nAlleles());
        }
        for (int j=0, n=data.nAlleles(); j<n; ++j) {
            out.print(space);
            out.print(data.alleles(j));
        }
        int n = data.nSamples() * data.nGenotypes();
        for (int j=0; j<n; ++j) {
            out.print(space);
            out.print(df.format(data.normData(j)));
        }
        out.println();
    }

    /**
     * Returns <code>true</code> if the output data will be written in
     * multi-allelic format and <code>false</code> otherwise.
     * @return <code>true</code> if the output data will be written in
     * multi-allelic format and <code>false</code> otherwise.
     */
    public boolean isMultiAllelic() {
        return isMultiAllelic;
    }

    /**
     * Returns the header line for a filename in BEAGLE .gprobs format
     * that corresponds to the samples in <code>data.samples()</code>.  The
     * header line will correspond to multi-allelic BEAGLE .gprobs format
     * if and only if <code>this.isMultiAllelic()==true</code>.
     * 
     * @param data a <code>GprobsData</code> instance.
     * @return the header line for a filename in BEAGLE .gprobs format
     * that corresponds to <code>data.samples()</code>.
     *
     * @throws NullPointerException if <code>data==null</code>.
     * @throws IllegalArgumentException if
     * <code>data.nGenotypes()!=3 && this.isMultiAllelic()==false</code>
     */
    public String getOutputHeaderLine(GprobsData data) {
        StringBuilder sb = new StringBuilder(10*data.nSamples());
        if (isMultiAllelic) {
            if (data.nSamples() > 0) {
                sb.append(data.sampleIds(0));
            }
            for (int j=1, n=data.nSamples(); j<n; ++j) {
                sb.append(space);
                sb.append(data.sampleIds(j));
            }
        }
        else {
            if (data.nGenotypes() != 3) {
                String s = "isMultiAllelic=false, but marker is multiAllelic: "
                        + data;
                throw new IllegalArgumentException(s);
            }
            assert data.nGenotypes()==3;
            sb.append("marker");
            sb.append(space);
            sb.append("alleleA");
            sb.append(space);
            sb.append("alleleB");
            for (int j=0, n=data.nSamples(); j<n; ++j) {
                for (int k=0; k<3; ++k) {
                    sb.append(space);
                    sb.append(data.sampleIds(j));
                }
            }
        }
        return sb.toString();
    }
}
