/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package beagleutil;

import java.util.*;
import blbutil.ImmutableSet;

/**
 * Class <code>Samples</code> is an immutable class that represents a filtered
 * and unfiltered list of sample identifiers.
 * 
 * @author Brian L Browning
 */
public class Samples {

    private boolean[] filter;
    private String[] unfilteredIds;
    private String[] filteredIds;

    /**
     * Constructs a new instance of </code>Samples</code>.
     * @param ids an array of sample identifiers.
     * @param exclusions a set of sample identifiers to exclude.
     * @throws NullPointerException if 
     * <code>ids==null || exclusions==null</code> or if any element of
     * <code>ids[j]==null</code> for <code>0 &le; j < ids.length</code>.
     * @throws IllegalArgumentException if
     * <code>ids[j].equals(ids[k])==false</code> for
     * <code>0 &le; j < k < ids.length</code>
     */
    public Samples(String[] ids, ImmutableSet<String> exclusions) {
        checkForDuplicatesAndNulls(ids);
        this.unfilteredIds = ids.clone();
        this.filter = new boolean[ids.length];
        List<String> filteredList = new ArrayList<String>(ids.length);
        for (int j=0; j<unfilteredIds.length; ++j) {
            if (exclusions.contains(unfilteredIds[j])) {
                filter[j] = true;
            }
            else {
                filteredList.add(unfilteredIds[j]);
            }
        }
        this.filteredIds = filteredList.toArray(new String[0]);
    }

    private void checkForDuplicatesAndNulls(String[] ids) {
        Set<String> set = new HashSet<String>(2*ids.length);
        for (String id : ids) {
            if (id==null) {
                String s = "Array contains a null identifier: "
                        + Arrays.toString(ids);
                throw new NullPointerException(s);
            }
            if (set.add(id)==false) {
                String s = "duplicate sample ID (" + id + ") in " +
                        Arrays.toString(ids);
                throw new IllegalArgumentException(s);
            }
        }
    }

    /**
     * Returns the number of samples before filtering.
     * @return the number of samples before filtering.
     */
    public int unfilteredSize() {
        return unfilteredIds.length;
    }

    /**
     * Returns the number of samples after filtering.
     * @return the number of samples after filtering.
     */
    public int filteredSize() {
        return filteredIds.length;
    }

    /**
     * Returns <code>true</code> if the specified unfiltered sample identifier
     * is excluded and <code>false</code> otherwise.
     * @param k an index of a unfiltered sample.
     * @return the <code>true</code> if the specified unfiltered sample 
     * identifier is excluded and <code>false</code> otherwise.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.unfilteredSize()</code>.
     */
    public boolean filter(int k) {
        return filter[k];
    }


    /**
     * Returns a boolean array of length <code>this.unfilteredSize()</code>
     * whose <code>true</code> entries identify the indices of excluded
     * samples in <code>this.unfilteredIds()</code>.
     * @return a boolean array of length <code>this.unfilteredSize()</code>
     * whose <code>true</code> entries identify the indices of excluded
     * samples in <code>this.unfilteredIds()</code>.
     */
    public boolean[] filter() {
        return filter.clone();
    }

    /**
     * Returns the specified unfiltered sample identifier.
     * @param k an index of a unfiltered sample.
     * @return the specified filtered sample identifier.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.unfilteredSize()</code>.
     */
    public String unfilteredIds(int k) {
        return unfilteredIds[k];
    }

    /**
     * Returns the specified filtered sample identifier.
     * @param k an index of a filtered sample.
     * @return the specified filtered sample identifier.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.filteredSize()</code>.
     */
    public String filteredIds(int k) {
        return filteredIds[k];
    }

    /**
     * Returns an array of unfiltered sample identifiers such that
     * <code>this.unfilteredsIds()[k].equals(this.unfilteredIds(k))</code>
     * for <code> 0 &le; k < this.unfilteredSize()</code>.
     *
     * @return an array of unfiltered sample identifiers.
     */
    public String[] unfilteredIds() {
        return unfilteredIds.clone();
    }

    /**
     * Returns an array of filtered sample identifiers such that
     * <code>this.filteredsIds()[k].equals(this.filteredIds(k))</code>
     * for <code> 0 &le; k < this.filteredSize()</code>.
     *
     * @return an array of filtered sample identifiers.
     */
    public String[] filteredIds() {
        return filteredIds.clone();
    }

    /**
     * Returns a string representation of <code>this</code>.
     * The exact details of the representation are unspecified and
     * subject to change.
     * @return a string representation of <code>this</code>.
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder(10*filter.length);
        if (filter.length > 0) {
            sb.append(unfilteredIds[0]);
            sb.append("=");
            sb.append(filter[0]);
        }
        for (int j=1; j<filter.length; ++j) {
            sb.append(" ");
            sb.append(unfilteredIds[j]);
            sb.append("=");
            sb.append(filter[j]);
        }
       return sb.toString();
    }
}
