/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package blbutil;

import java.io.*;
import java.util.NoSuchElementException;

/**
 * An iterator for text files that iterates through the lines of an
 * input stream.  The <code>next</code> method returns <code>String</code>
 * instances that correspond to lines of the input stream.
 * If an <code>IOException</code> is thrown while reading a file, the
 * <code>IOException</code> is trapped, an appropriate error message is
 * written to standard out, and the Java Virtual Machine is terminated.
 *
 * @author Brian L Browning
 */
public class InputStreamIterator implements FileIterator<String> {

    private BufferedReader in = null;
    private String next = null;
    private final InputStream inputStream;

    /**
     * Creates an <code>InputStreamIterator</code> to iterate through
     * the lines of the specified input stream.
     *
     * @param inputStream input stream that will be read one line at a time.
     */
    public InputStreamIterator(InputStream inputStream) {
        this.inputStream = inputStream;
        try {
            in = new BufferedReader(new InputStreamReader(inputStream));
            next = in.readLine();
        }
        catch(IOException e) {
            Utilities.exit("Error reading " + inputStream, e);
        }      
    }
    
    public boolean hasNext() {
        return (next != null);
    }

    public String next() {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        String current = next;
        try {
            next = in.readLine();
        }
        catch (IOException e) {
            Utilities.exit("Error reading " + in, e);
        }
        return current;
    }

    public void remove() {
        String s = "remove() is not supported by LineIterator";
        throw new UnsupportedOperationException(s);
    }

    public void close() {
        try {
            in.close();
        }
        catch (IOException e) {
            Utilities.exit("Error closing " + in, e);
        }
        next=null;
    }
    
    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer("[input stream=" + inputStream);
        sb.append("\nnext = " + next);
        sb.append("\nBufferedReader = " + in);
        return sb.toString();
    }

    /**
     * Returns an <code>FileIterator<String></code> for the specified
     * text file.  If the filename ends in ".gz", the file is assumed
     * to be GZIP-compressed and will automatically be uncompressed.
     * If an <code>IOException</code> or <code>FileNotFoundException</code>
     * is thrown, the virtual machine will exit with an error message.
     * @param file a text file.
     * @return a <code>FileIterator<String></code> for the specified
     * text file.
     *
     * @throws NullPointerException if <code>file==null</code>.
     */
    public static InputStreamIterator fromGzipFile(File file) {
        try {
            InputStream is = new FileInputStream(file);
            if (file.getName().endsWith(".gz")) {
                return new InputStreamIterator(
                        new java.util.zip.GZIPInputStream(is));
            }
            else {
                return new InputStreamIterator(is);
            }
        }
        catch(FileNotFoundException e) {
            Utilities.exit("Error opening " + file, e);
        }
        catch(IOException e) {
            Utilities.exit("Error reading " + file, e);
        }
        assert false;
        return null;
    }

    /**
     * Returns an <code>FileIterator<String></code> for the specified
     * text file.  If the filename ends in ".gz", the file is assumed
     * to be GZIP-compressed and will automatically be uncompressed.
     * If an <code>IOException</code> or <code>FileNotFoundException</code>
     * is thrown, the virtual machine will exit with an error message.
     * @param filename name of a text file.
     * @return a <code>FileIterator<String></code> for the specified
     * text file.
     *
     * @throws NullPointerException if <code>file==null</code>.
     */
    public static InputStreamIterator fromGzipFile(String filename) {
        return fromGzipFile(new File(filename));
    }

    /**
     * Returns an <code>FileIterator<String></code> for the specified
     * text file.  If an <code>IOException</code> or
     * <code>FileNotFoundException</code> is thrown, the virtual machine
     * will exit with an error message.
     * @param file a text file.
     * @return a <code>FileIterator<String></code> for the specified
     * text file.
     *
     * @throws NullPointerException if <code>file==null</code>.
     */
    public static InputStreamIterator fromTextFile(File file) {
        try {
            return new InputStreamIterator(new FileInputStream(file));
        }
        catch(FileNotFoundException e) {
            Utilities.exit("Error opening " + file, e);
        }
        catch(IOException e) {
            Utilities.exit("Error reading " + file, e);
        }
        assert false;
        return null;
    }

    /**
     * Returns an <code>FileIterator<String></code> for the specified
     * text file.  If an <code>IOException</code> or
     * <code>FileNotFoundException</code> is thrown, the virtual machine
     * will exit with an error message.
     * @param filename name of a text file.
     * @return an <code>FileIterator<String></code> for the specified
     * text file.
     *
     * @throws NullPointerException if <code>filename==null</code>.
     */
    public static InputStreamIterator fromTextFile(String filename) {
        return fromTextFile(new File(filename));
    }
}
