/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package blbutil;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.zip.*;

/**
 * Class <code>Utilities</code> contains general utility methods.
 *
 * @author Brian L Browning, except where noted
 */
public class Utilities {

    private static final int BUFFER_SIZE = 2*(1<<20);

    private Utilities() {
        // private constructor to prevent instantiation
    }

    /**
     * Returns a string representing the current local time.  The
     * exact details of the representation are unspecified and
     * subject to change.
     * 
     * @return a string representing the current local time.
     */
    public static String timeStamp() {
        Date now = new Date();
        SimpleDateFormat sdf =
                new SimpleDateFormat("hh:mm a z 'on' dd MMM yyyy");
        return sdf.format(now);
    }

    /**
     * Returns the number of null entries in the specified array.
     * @param array an array
     * @return the number of null entries in the specified array.
     * @throws NullPointerException if <code>array==null</code>
     */
    public static int countNull(Object[] array) {
        int cnt = 0;
        for (Object o : array) {
            if (o==null) {
                ++cnt;
            }
        }
        return cnt;
    }

    /**
     * Returns the number of true entries in the specified array.
     * @param array a boolean array
     * @return the number of true entries in the specified array.
     * @throws NullPointerException if <code>array==null</code>
     */
    public static int countTrue(boolean[] array) {
        int cnt = 0;
        for (boolean b : array) {
            if (b) {
                ++cnt;
            }
        }
        return cnt;
    }

    /**
     * Returns the first non-blank line of the specified file.
     * @param file a text file
     * @return the first line of the specified file.
     * @throws NullPointerException if <code>f==null</code>.
     * @throws IllegalArgumentException if the specified file has
     * no blank lines.
     */
    public static String readFirstNonBlankLine(File file) {
        FileIterator<String> it = InputStreamIterator.fromGzipFile(file);
        String firstLine = null;
        while (it.hasNext() && firstLine == null) {
            String candidateLine = it.next().trim();
            if (candidateLine.length() > 0) {
                firstLine = candidateLine;
            }
        }
        it.close();
        if (firstLine==null) {
            String s = "ERROR: empty file: " + file;
            throw new IllegalArgumentException(s);
        }
        return firstLine;
    }

    /**
     * Returns a set of identifiers found in a single column text file.
     * The empty set is returned if <code>exclusionFile==null</code>.
     * @param file a text file with a single column of data
     * @return a set of identifiers found in a single column text file.
     * The empty set is returned if <code>exclusionFile==null</code>.
     * @throws IllegalArgumentException if any line of the specified
     * file contains more than one white-space delimited fields.
     */
    public static Set<String> getIdSet(File file) {
        Set<String> exclusions = new HashSet<String>();
        if (file != null) {
            FileIterator<String> it = InputStreamIterator.fromGzipFile(file);
            while (it.hasNext()) {
                String id = it.next().trim();
                int length = id.length();
                if (length > 0) {
                    if (StringUtil.countFields(id) > 1) {
                        String s = "Error: >1 excluded identifier in line: "+id;
                        throw new IllegalArgumentException(s);
                    }
                    exclusions.add(id);
                }
            }
            it.close();
        }
        return exclusions;
    }

    /**
     * Returns a set that contains all non-empty lines of the
     * specified text file containing a single column of data.
     * Any white space at the beginning or end of each line is removed.
     * @param filename a text file
     * @return a set that contains all non-empty lines of the
     * specified text file.
     * 
     * @throws IllegalArgumentException if any line of the specified text file
     * contains two or more white-space delimited fields.
     * @throws NullPointerException if <code>filename==null</code>.
     */
    public static Set<String> column2Set(String filename) {
        FileIterator<String> it = InputStreamIterator.fromGzipFile(filename);
        Set<String> set = new HashSet<String>(5000);
        while (it.hasNext()) {
            String line = it.next().trim();
            if (line.length() > 0) {
                if (StringUtil.countFields(line) != 1) {
                    String s = "Error: >1 field on line: " + line;
                    throw new IllegalArgumentException(s);
                }
                set.add(line);
            }
        }
        return set;
    }

    /*
     * Prints the specified string to both the specified PrintWriter and
     * to standard out.  Line separator characters are not appended to the
     * specified string.
     *
     * @param out a PrintWriter used to print the specified string.
     * @param s a string to be printed.
     */
    public static void duoPrint(PrintWriter out, String s) {
        System.out.print(s);
        out.print(s);
    }

    /**
     * Prints the delimited fields to the specified PrintWriter followed
     * by a new line character.
     * 
     * @param fields fields to be printed.
     * @param delimiter a delimiter for the fields
     * @param out an output printwriter
     * @throws NullPointerException if <code>fields==null</code>.
     */
    public static void printLine(String[] fields, String delimiter,
            PrintWriter out) {
        if (fields.length > 0) {
            out.print(fields[0]);
        }
        for (int j=1; j<fields.length; ++j) {
            out.print(delimiter);
            out.print(fields[j]);
        }
        out.println();
    }
    
    /**
     * Prints the specified string, the specified exception, and a 
     * stack trace to standard out and terminates the java interpreter.
     *
     * @param s a message to be printed to standard err.
     * @param e an exception to be printed to standard err.
     */
    public static void exit(String s, Exception e) {
        e.printStackTrace();
        System.err.println(s);
        System.err.println("terminating program.");
        System.exit(1);
    }  

    /**
     * Returns the number of nonempty lines in the specified file.
     * A nonempty line contains at least one character that is
     * not white space.  Files whose names end in ".gz" are 
     * assumed to be GZIP compressed.
     * @param file a text file.
     * @return the number of nonempty lines in the specified file.
     * 
     * @throws NullPointerException if <code>file==null</code>.
     */
    public static int countNonBlankLines(File file) {
        int cnt = 0;
        FileIterator<String> it = InputStreamIterator.fromGzipFile(file);
        while (it.hasNext()) {
            String line = it.next().trim();
            if (line.length() > 0) {
                ++cnt;
            }
        }
        it.close();
        return cnt;
    }
    
    /**
     * Copies the specified text file.
     *
     * @param from text file to copy
     * @param to copy of text file to be created.
     *
     * @throws NullPointerException if 
     * <code>from==null || to==null</code>.
     */
    public static void copyFile(String from, String to) {
        FileIterator<String> it = InputStreamIterator.fromTextFile(from);
        PrintWriter out = Utilities.createPrintWriter(to);
        while (it.hasNext()) {
            out.println(it.next());
        }
        out.close();
        it.close();
    }    
    
   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified filename.  Any existing file corresponding 
    * with the specified filename will be deleted.  If the file 
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param outFile the name of the file to be opened for output. 
    * 
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    */    
    public static PrintWriter createPrintWriter(String outFile) {
        return createPrintWriter(new File(outFile), false);
    }

   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified filename.  If <code>append==false</code>
    * any existing file corresponding 
    * with the specified filename will be deleted.  If the file 
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param outFile the name of the file to be opened for output. 
    * @param append if <code>true</code>, then data will be written to the 
    * end of the file rather than the beginning.
    * 
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    */    
    public static PrintWriter createPrintWriter(String outFile, boolean append) {
        return createPrintWriter(new File(outFile), append);
    }    
    
   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified filename.  Any existing file corresponding 
    * with the specified filename will be deleted.  If the file 
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param file the file to be opened for output. 
    * 
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    */    
    public static PrintWriter createPrintWriter(File file) {
        return createPrintWriter(file, false);
    }

   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified filename.  If <code>append==false</code> 
    * any existing file corresponding 
    * with the specified filename will be deleted.  If the file 
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param file the file to be opened for output. 
    * @param append if <code>true</code>, then data will be written to the 
    * end of the file rather than the beginning. 
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    */    
    public static PrintWriter createPrintWriter(File file, boolean append) {
        return createPrintWriter(file, append, BUFFER_SIZE);
    }

   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified filename.  If <code>append==false</code>
    * any existing file corresponding
    * with the specified filename will be deleted.  If the file
    * cannot be opened, an error message is printed and the
    * java interpreter exits.
    *
    * @param file the file to be opened for output.
    * @param append if <code>true</code>, then data will be written to the
    * end of the file rather than the beginning.
    * @param size the buffer size
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    * @throws IllegalArgumentException if <code>size &le; 0</code>.
    */
    public static PrintWriter createPrintWriter(File file, boolean append,
            int size) {
        PrintWriter out = null;
        try {
            out = new PrintWriter(
                    new BufferedWriter(new FileWriter(file, append), size));
        }
        catch (IOException e) {
            Utilities.exit("Error opening "+ file, e);
        }
        return out;
    }

   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.  The resulting file will be compressed using
    * the GZIP compression algorithm.  Any existing file corresponding
    * to the specified file will be deleted.  If the file
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param file the file to be opened for output. 
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    */    
    public static PrintWriter createGzipPrintWriter(File file) {
        return createGzipPrintWriter(file, BUFFER_SIZE);
    }

   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.  The resulting file will be compressed using
    * the GZIP compression algorithm.  Any existing file corresponding
    * to the specified file will be deleted.  If the file
    * cannot be opened, an error message is printed and the
    * java interpreter exits.
    *
    * @param file the file to be opened for output.
    * @param size the buffer size.
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    * @throws IllegalArgumentException if <code>size &le; 0</code>.
    */
    public static PrintWriter createGzipPrintWriter(File file, int size) {
        PrintWriter out = null;
        try {
            out = new PrintWriter(
                    new GZIPOutputStream(
                    new BufferedOutputStream(new FileOutputStream(file)), size));
        }
        catch (IOException e) {
            Utilities.exit("Error opening "+ file, e);
        }
        return out;
    }

   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.  The resulting file will be compressed using
    * the GZIP compression algorithm.  Any existing file corresponding
    * to the specified file will be deleted.  If the file
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param filename the name of the file to be opened.
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    * 
    * @throws NullPointerException if <code>filename==null</code>.
    */    
    public static PrintWriter createGzipPrintWriter(String filename) {
        return createGzipPrintWriter(new File(filename));
    }

    /**
     * Returns a <code>java.io.DataOutputStream</code> corresponding to
     * the specified file.  Any existing file corresponding to the 
     * <code>File</code> object will be deleted.   If the file cannot be
     * opened, an error message is printed and the java interpreter exits.
     *
     * @param file the file to be opened for output.  
     *
     * @return a <code>java.io.DataOutputStream</code> corresponding to
     * the specified file.
     */
    public static DataOutputStream createDataOutputStream(File file) {
        return createDataOutputStream(file, false);
    }

    /**
     * Returns a <code>java.io.DataOutputStream</code> corresponding to
     * the specified file.  Any existing file corresponding to the 
     * <code>File</code> object will be deleted.   If the file cannot be
     * opened, an error message is printed and the java interpreter exits.
     *
     * @param file the file to be opened for output.
     * @param append if <code>true</code>, then bytes will be written to the 
     * end of the file rather than the beginning.
  
     *
     * @return a <code>java.io.DataOutputStream</code> corresponding to
     * the specified file.
     */
    public static DataOutputStream createDataOutputStream(File file,
            boolean append) {
        int size = BUFFER_SIZE;
        OutputStream os = null;
        try {
            os = new FileOutputStream(file, append);
        }
        catch (FileNotFoundException e) {
            Utilities.exit("Error opening " + file, e);
        }
        DataOutputStream out = new DataOutputStream(
                new BufferedOutputStream(os, size));
        return out;
    }    
    
    /**
     * Closes the specified output stream.  If the output stream cannot be
     * closed, an error message is printed and the java interpreter exits.
     * 
     * @param os a output stream.
     */    
    public static void closeOutputStream(OutputStream os) {
        try {
            os.close();
        }
        catch (IOException e) {
            Utilities.exit("Error closing output stream" + os, e);            
        }        
    }
        
    /**
     * Returns a data input stream reading from the specified file.  If the
     * input stream cannot be opened, an error message is printed and the
     * java interpreter exits.
     * 
     * @param file a binary file.
     *
     * @return a data input stream reading from the specified file.
     */
    public static DataInputStream createDataInputStream(File file) {
        int size = BUFFER_SIZE;
        DataInputStream is = null;
        try {
            is = new DataInputStream(new BufferedInputStream(
                    new FileInputStream(file), size));
        }
        catch (FileNotFoundException e) {
            Utilities.exit("Error opening " + file, e);
        }    
        return is;
    }  
  
    /*
     * Closes the specified data input stream.  If the data input
     * stream cannot be closed, an error message is printed and the
     * java interpreter exits.
     * 
     * @param is a data input stream.
     */
    public static void closeDataInputStream(DataInputStream is) {
        try {
            is.close();
        }
        catch (IOException e) {
            Utilities.exit("Error closing input stream: " + is, e);
        }    
    }
    
    /**
     * Returns an array of empty temporary files.  The temporary files 
     * will be deleted when the java virtual machine exits.
     * 
     * @param size the number of temporary files that will be created.
     * @param prefix the prefix string to be used in generating each 
     * file's name; must be at least three characters long.
     * 
     * @return an array of empty temporary files. The temporary files 
     * will be deleted when the java virtual machine exits.
     * 
     * @throws IllegalArgumentException if the <code>prefix</code> 
     * argument contains fewer than three characters.
     * @throws NullPointerException if <code>prefix==null</code>.
     * @throws NegativeArraySizeException if <code>size < 0</code>.
     */
    public static File[] getTempFileArray(int size, String prefix) {
        if (prefix==null) {
            throw new NullPointerException("prefix==null");
        }
        File[] fa = new File[size];
        for (int j=0; j<fa.length; ++j) {
            fa[j] = Utilities.tempFile(prefix);
        }
        return fa;
    }
    
   /**
    * Returns a temporary <code>File</code> that will be deleted when 
    * the java virtual machine exits.
    * 
    * @param prefix the prefix string to be used in generating the file's 
    * name; must be at least three characters long.
    *
    * @return a <code>File</code> with an abstract pathname denoting a 
    * newly created empty file.
    * 
    * @throws IllegalArgumentException if the <code>prefix</code> 
    * argument contains fewer than three characters.
    */
    public static File tempFile(String prefix) {
       File tempFile = null;
       try {
            tempFile = File.createTempFile(prefix, null);
            tempFile.deleteOnExit();
       } 
       catch (Exception e) {
           System.out.println("Exception thrown by createTempFile: " + e);
           System.exit(1);
       }
       return tempFile;
    }    
    

    /**
     * Returns an <code>int</code> array obtained from parsing
     * the elements of the specified <code>String</code> array.
     * 
     * @param a <code>String</code> array whose elements are
     * parsable integers.
     * @return an <code>int</code> array obtained from parsing the 
     * elements of the specified <code>String</code> array.
     * @throws NumberFormatException if the specified array has a 
     * <code>String</code> element that does not contain a parsable
     * integer.
     */
    public static int[] toIntArray(String[] a) {
        int[] intArray = new int[a.length];
        for (int j=0; j<intArray.length; ++j) {
            intArray[j] = Integer.parseInt(a[j]);
        }
        return intArray;    
    }
    
    /**
     * Returns an <code>int</code> array obtained from parsing
     * the specified elements of the specified <code>String</code> 
     * array .
     * 
     * @param a <code>String</code> array whose elements are
     * parsable integers.
     * @param start the index of the first array entry to 
     * be parsed.
     * @param end one more than the index of the last array entry
     * to be parsed.
     * @return an <code>int</code> array of length <code>end - start</code>
     * obtained from parsing the specified elements of the specified 
     * <code>String</code> array.
     * @throws NumberFormatException if the specified array has a 
     * <code>String</code> element that does not contain a parsable
     * integer.
     * @throws ArrayIndexOutOfBoundsException if 
     * <code> start < 0 || end &ge; a.length</code>.
     * @throws NegativeArraySizeException if <code>end > start</code>.
     */
    public static int[] toIntArray(String[] a, final int start, final int end) {
        int[] intArray = new int[end - start];
        for (int j=start; j<end; ++j) {
            intArray[j-start] = Integer.parseInt(a[j]);
        }
        return intArray;            
    }

    /**
     * Returns a map with one (key, value) pair for each string element
     * of the specified array that contains the specified separator.
     * If an array element <code>s</code> of the specified array contains 
     * the specific separator character then the key is
     * <code> s.substring(0, s.indexOf('sep'))</code>
     * and the value is <code>s.substring(s.indexOf(s.indexOf(sep) + 1)</code>.
     * Thus if the first occurrence of the specified separator is the 
     * last character of the string, the map value will be the
     * empty string ("").
     * <p/>
     * 
     * If the array element <code>s</code> of the specified array does not 
     * contain the specified separator character then the entry is ignored.
     * 
     * @param args a string array.
     * @param sep a separator character used to create a (key, value) pair 
     * from a string of the form <code>key + sep + value</code>.
     * @return a <code>java.util.Map<String, String></code> with one
     *        (key, value) pair for each element of the specified
     *        string array with the form <code>key + sep + value</code>.
     * @throws NullPointerException if <code>args==null</code>.
     * @throws IllegalArgumentException if any two elements of the 
     * specified string array have the same key.
     */
    public static Map<String,String> argsToMap(String[] args, char sep) {
        Map<String,String> argMap = new HashMap<String,String>();
        for (String arg : args) {
            String value = null;
            int index = arg.indexOf(sep);
            if (index != -1) {
                String key = arg.substring(0, index);
                value = arg.substring(index + 1);
                if (argMap.containsKey(key)) {
                    String s = "duplicate arguments: " + key;
                    throw new IllegalArgumentException(s);
                }
                argMap.put(key, value);                
            }
        }
        return argMap;
    }

    /**
     * Returns a multimap whose values are <code>String[]</code> arrays
     * which contain all values associated with the key.  The order of
     * elements in the array corresponding to a key is not required to
     * correspond to the order of the elements in the <code>args</code>
     * array.
     * 
     * If an array element <code>s</code> of the specified array contains 
     * the specific separator character then the key is
     * <code> s.substring(0, s.indexOf('sep'))</code>
     * and the value is <code>s.substring(s.indexOf(s.indexOf(sep) + 1)</code>.
     * Thus if the first occurrence of the specified separator is the
     * last character of the string, an element of the map value will be the
     * empty string ("").
     * <p/>
     * 
     * If the array element <code>s</code> of the specified array does not 
     * contain the specified separator character then the entry is ignored.
     * 
     * @param args a string array.
     * @param sep a separator character delimiting a (key, value) pair 
     * from a string of the form <code>key + sep + value</code>.
     * @return a <code>java.util.Map<String, String[]></code> whose keyset
     *        constains all keys in the specified string array.  The value
     *        associated with each key in the map keyset is the 
     *        string array of all values associated with that key in the 
     *        specified string array.
     * @throws NullPointerException if <code>args==null</code>.
     * @throws IllegalArgumentException if any two elements of the 
     * specified string array donote the same the same (key, value) pair.
     */
    public static Map<String,String[]> argsToMultimap(String[] args, char sep) {
        Map<String,Collection<String>> map =
                new HashMap<String,Collection<String>>();
        for (String arg : args) {
            String value = null;
            int index = arg.indexOf(sep);
            if (index != -1) {
                String key = arg.substring(0, index);
                value = arg.substring(index + 1);
                if (map.containsKey(key)) {
                    Collection<String> c = map.get(key);
                    if (c.contains(value)) {
                        String s = "duplicate arguments: " + arg;
                        throw new IllegalArgumentException(s);
                    }
                    else {
                        c.add(value);
                    }
                }
                else {
                    Collection<String> c = new ArrayList<String>();
                    c.add(value);
                    map.put(key, c);
                }
            }
        }
        Map<String,String[]> argMap = new HashMap<String,String[]>();
        String[] keySet = map.keySet().toArray(new String[0]);
        for (String key : keySet) {
            Collection<String> c = map.get(key);
            argMap.put(key, c.toArray(new String[0]));
            map.remove(key);
        }
        return argMap;
    }
    
    /**
     * Returns the string entries that do not contain the specified 
     * separator character as a list.  The elements of the returned list
     * will have the same ordering as in the specified string array.
     *
     * @param args a string array.
     * @param sep a separator character.  Elements of the string array
     * that contains the separator character are ignored.
     *
     * @return the string entries that do not contain the specified 
     * separator character as a list.
     *
     * @throws NullPointerException if <code>args==null</code>.
     */
    public static List<String> argsToList(String[] args, char sep) {
        List<String> argList = new ArrayList<String>();
        for (String arg : args) {
            int index = arg.indexOf(sep);
            if (index==-1) {
                argList.add(arg);
            }
        }
        return argList;
    }

     /**
     * Returns a string representation of the specified elapsed time.
     *
     * @param milliseconds the elapsed time in milliseconds
     *
     * @return a string representation of the specified elapsed time.
     */
    public static String elapsedTime(long milliseconds) {
        long seconds = Math.round(milliseconds /1000.0);
        StringBuilder sb = new StringBuilder();
        if (seconds >= 3600) {
            sb.append(seconds / 3600);
            sb.append(" hours ");
            seconds %= 3600;
 
        }
        if (seconds >= 60) { 
            sb.append(seconds / 60);
            sb.append(" minutes ");
            seconds %= 60;
        }
        sb.append(seconds);
        sb.append(" seconds");
        return sb.toString();
    } 
    
    /**
     * Returns the floating-point value adjacent to <code>d</code> in
     * the direction of positive infinity.  This method is
     * semantically equivalent to <code>nextAfter(d,
     * Double.POSITIVE_INFINITY)</code>; however, a <code>nextUp</code>
     * implementation may run faster than its equivalent
     * <code>nextAfter</code> call.
     *
     * <p>Special Cases:
     * <ul>
     * <li> If the argument is NaN, the result is NaN.
     *
     * <li> If the argument is positive infinity, the result is
     * positive infinity.
     *
     * <li> If the argument is zero, the result is
     * <code>Double.MIN_VALUE</code>
     *
     * </ul>
     *
     * @param d  starting floating-point value
     * @return The adjacent floating-point value closer to positive
     * infinity.
     * @author Joseph D. Darcy
     * copied from sun.misc.FpUtils
     */
    public static double nextUp(double d) {
        if (Double.isNaN(d) || d == Double.POSITIVE_INFINITY) {
            return d;
        } else {
            d += 0.0d;
            return Double.longBitsToDouble(Double.doubleToRawLongBits(d) +
                    ((d >= 0.0d) ? +1L : -1L));
        }
    }

    /**
     * Returns the floating-point value adjacent to <code>d</code> in
     * the direction of negative infinity.  This method is
     * semantically equivalent to <code>Math.nextAfter(d,
     * Double.NEGATIVE_INFINITY)</code>; however, a <code>nextDown</code>
     * implementation may run faster than its equivalent
     * <code>nextAfter</code> call.
     *
     * <p>Special Cases:
     * <ul>
     * <li> If the argument is NaN, the result is NaN.
     *
     * <li> If the argument is negative infinity, the result is
     * negative infinity.
     *
     * <li> If the argument is zero, the result is
     * <code>-Double.MIN_VALUE</code>
     *
     * </ul>
     *
     * @param d  starting floating-point value
     * @return The adjacent floating-point value closer to negative
     * infinity.
     * @author Joseph D. Darcy
     * copied from sun.misc.FpUtils
     */
    public static double nextDown(double d) {
        if (Double.isNaN(d) || d == Double.NEGATIVE_INFINITY) {
            return d;
        } else {
            if (d == 0.0) {
                return -Double.MIN_VALUE;
            } else {
                return Double.longBitsToDouble(Double.doubleToRawLongBits(d) +
                        ((d > 0.0d) ? -1L : +1L));
            }
        }
    }

}

