/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import java.io.*;
import java.text.DecimalFormat;
import java.util.*;
import blbutil.*;
import beagleutil.SNP_HWE;
import beagleutil.GprobsData;
import beagleutil.GprobsPrinter;

/**
 * Driver class for the BEAGLECALL program.
 * @author Brian L Browning
 */
public class CallMain {
    
    private static final String nl = CallingParameters.newLine;
    private static final String tab = "\t";
    private static final String version = "BeagleCall version 1.0.1 (15 Nov 2010)";
    private static final String shortHelp = version + nl
            + "Enter \"java -jar beaglecall.jar\" for summary of command " +
            "line arguments.";

    private static final DecimalFormat dfsig2 = new DecimalFormat("0.00E0");
    private static final DecimalFormat df4 = new DecimalFormat("0.0000");
    private static final DecimalFormat df5 = new DecimalFormat("0.#####");

    private final boolean isMultiAllelic = false;
    private final boolean isCompressed = true;
    private CallingParameters parameters;

    private CallingDataIterator dataIt;
    private PrintWriter logOut;
    private PrintWriter hweOut;
    private PrintWriter missingOut;
    private PrintWriter excludeOut;
    private PrintWriter markersOut;
    private PrintWriter[] modelOut;
    private GprobsPrinter[] gprobsPrinter;
    private GprobsPrinter[] likePrinter;
    private PrintWriter[] gprobsOut;
    private PrintWriter[] likeOut;

    private int nSignalsMarkers = 0; // number of markers in first .signals file
    private int nGprobsMarkers = 0;  // number of markers in first .gprobs file
    private int priorExclusionCnt = 0;
    private int hweExcludeCnt = 0;
    private int missExcludeCnt = 0;
    private int runtimeExclusionCnt = 0;// from Hardy-Weinberg OR missing filter
    private int readMarkerCnt = 0;      // doesn't count prior exclusions
    private int nOutputMarkers = 0;

    /**
     * Entry point to BeagleCall program.  See .pdf documentation for
     * information on program arguments.
     * @param args command line arguments
     */
    public static void main(String[] args) {
	Locale.setDefault(Locale.US);
	CallMain cm = new CallMain(args);
    }

    /*
     * Constructs a <code>CallMain</code> instance that runs
     * the beaglecall program using the specified arguments.
     * Private constructor prevents public instantiation.
     *
     * @param args command line arguments
     */
    private CallMain(String[] args) {
        long startMillis = System.currentTimeMillis();
        System.out.println(shortHelp);
        if (args.length==0) {
            System.out.println(nl + CallingParameters.usage());
            System.exit(0);
        }
        initializeClassFields(args);

        logOut.println(shortHelp);
        Utilities.duoPrint(logOut, "Start time: " + Utilities.timeStamp() + nl);
        Utilities.duoPrint(logOut, beaglecallCommandLine(args));
        processMarkers();
        printSummary(startMillis);
        
        String[] beagleArgs = beagleArgs();
        if (false) {
            Utilities.duoPrint(logOut, nl);
            closeResources();
            //phaser.PhaseMain.main(beagleArgs);
        }
        else {
            Utilities.duoPrint(logOut, beagleCommandLine(beagleArgs));
            closeResources();
        }
    }

    private void initializeClassFields(String[] args) {
        CallingParameters p = new CallingParameters(args);
        checkParameters(p);
        String prefix = p.out();
        File[] sigFiles = p.signals();
        File[] gprobsFiles = p.gprobs();
        assert gprobsFiles.length > 0;

        this.parameters = p;
        this.modelOut = new PrintWriter[sigFiles.length];
        this.gprobsPrinter = new GprobsPrinter[sigFiles.length];
        this.likePrinter = new GprobsPrinter[sigFiles.length];
        String modelHeaderLine = ModelMarker.modelHeaderLine();
        for (int j=0; j<this.modelOut.length; ++j) {
            String s = prefix + "." + sigFiles[j].getName() + ".model";
            modelOut[j] = Utilities.createPrintWriter(s);
            modelOut[j].println(modelHeaderLine);
            gprobsPrinter[j] = new GprobsPrinter(df5, isMultiAllelic);
            likePrinter[j] = new GprobsPrinter(df5, isMultiAllelic);
        }
        this.logOut = Utilities.createPrintWriter(prefix + ".log");
        this.gprobsOut = createPrintWriters(prefix, sigFiles, ".tmpgp.gz");
        this.likeOut = createPrintWriters(prefix, sigFiles, ".like.gz");
        this.hweOut = Utilities.createPrintWriter(prefix + ".hwe");
        this.missingOut = Utilities.createPrintWriter(prefix + ".miss");
        this.markersOut = Utilities.createPrintWriter(prefix + ".markers");
        ImmutableSet<String> sampleExcl = readSet(p.excludesamples());
        ImmutableSet<String> markerExcl = readSet(p.excludemarkers());
        this.excludeOut = Utilities.createPrintWriter(prefix + ".markers.excl");
        this.nSignalsMarkers = Utilities.countNonBlankLines(sigFiles[0]) - 1;
        writeExclusions(gprobsFiles[0], markerExcl, this.excludeOut);
        this.dataIt = new CallingDataIterator(sigFiles, gprobsFiles, sampleExcl,
                markerExcl, isMultiAllelic);
    }

    private String beaglecallCommandLine(String[] args) {
        return commandLine("Command line: ", args, "beaglecall.jar", true);
    }

    private String beagleCommandLine(String[] args) {
        String prefix = "To finish, you must run BEAGLE.  " +
                "The suggested command line is:" + nl + nl;
        return commandLine(prefix, args, "beagle.jar", false);
    }

    /**
     * Returns a string representation of the command line arguments.
     *
     * @param args command line arguments.
     * @returns a string representation of the command line arguments.
     */
    private static String commandLine(String prefix, String[] args,
            String program, boolean calculateXmx) {
        StringBuilder sb = new StringBuilder(args.length*20);
        long maxMemory = Runtime.getRuntime().maxMemory();
        sb.append(nl);
        sb.append(prefix);
        sb.append("java -Xmx");
        if (maxMemory!=Long.MAX_VALUE && calculateXmx) {
            long maxMb = maxMemory / (1024*1024);
            sb.append(maxMb + "m");
        }
        else {
            sb.append("<Mb>m");
        }
        sb.append(" -jar ");
        sb.append(program);
        sb.append(" \\");
        sb.append(nl);
        for (int j = 0; j < args.length; ++j) {
            sb.append("  ");
            sb.append(args[j]);
            if (j < args.length - 1) {
                sb.append(" \\");
            }
            sb.append(nl);
        }
        return sb.toString();
    }

    private void checkParameters(CallingParameters p) {
        File f = new File(p.out());
        if (f.isDirectory()) {
            String s = "ERROR: \"out\" parameters cannot be a directory: "
                    + p.out();
            throw new IllegalArgumentException(s);
        }
        int nSignalFiles = p.signals().length;
        int nGprobsFiles = p.gprobs().length;
        if (nSignalFiles != nGprobsFiles) {
            String s = "unequal number of \"signals\" and \"gprobs\" files: "
                    + nSignalFiles + " != " + nGprobsFiles;
            throw new IllegalArgumentException(s);
        }
        if (nSignalFiles == 0) {
            throw new IllegalArgumentException("missing \"signals\" files.");
        }
    }

    private ImmutableSet<String> readSet(File f) {
        Set<String> idSet = Utilities.getIdSet(f);
        return new ImmutableSet<String>(idSet);
    }

    private void writeExclusions(File gprobsFile,
            ImmutableSet<String> exclusions, PrintWriter excludeOut) {
        boolean firstLine = true;
        FileIterator<String> fit = InputStreamIterator.fromGzipFile(gprobsFile);
        while (fit.hasNext()) {
            String line = fit.next().trim();
            if (line.length() > 0) {
                if (firstLine) {    // skip header line;
                    firstLine = false;
                }
                else {
                    ++nGprobsMarkers;
                    String[] fields = StringUtil.getFields(line, 2);
                    if (exclusions.contains(fields[0])) {
                        ++priorExclusionCnt;
                        excludeOut.println(fields[0]);
                    }
                }
            }
        }
        fit.close();
    }

    private void processMarkers() {
        while (dataIt.hasNext()) {
            CallingData cd = dataIt.next();
            if (!isExcluded(cd)) {
                writeMarkersFileLine(cd);
                double[] freq = genotypeFrequencies(cd);
                for (int j=0, n=cd.nCohorts(); j<n; ++j) {
                    ModelMarker mm = new ModelMarker(cd.signals(j), cd.gprobs(j),
                            parameters, freq);
                    modelOut[j].println(mm.model());
                    gprobsPrinter[j].print(mm.postGprobs(), gprobsOut[j]);
                    likePrinter[j].print(mm.postLike(), likeOut[j]);
                }
            }
            printScreenUpdate(++readMarkerCnt);
        }
    }

    private void printScreenUpdate(int markerIndex) {
        if (markerIndex==0) {
            System.out.print(nl + "Beginning analysis   ");
        }
        else if ((markerIndex % 500)==0) {
            if ((markerIndex % 10000)==0) {
                System.out.print(nl + "Finished marker " + markerIndex);
            }
            else {
                System.out.print(" .");
            }
        }
    }

    private void writeMarkersFileLine(CallingData cd) {
        markersOut.print(cd.markerId());
        markersOut.print(tab);
        markersOut.print(++nOutputMarkers);
        for (int j=0, n=cd.nAlleles(); j<n; ++j) {
            markersOut.print(tab);
            markersOut.print(cd.alleles(j));
        }
        markersOut.println();
    }

    private boolean isExcluded(CallingData cd) {
        boolean hweExclude = hweFilter(cd);
        boolean missingExclude = missingFilter(cd);
        boolean isExcluded = hweExclude || missingExclude;
        if (hweExclude) {
            ++hweExcludeCnt;
        }
        if (missingExclude) {
            ++missExcludeCnt;
        }
        if (isExcluded) {
            ++runtimeExclusionCnt;
            excludeOut.println(cd.markerId());
        }
        return isExcluded;
    }

    /**
     * Return <code>true</code> if marker is excluded by HWE filter and 
     * <code>false</code> otherwise, and writes per-cohort and per-study 
     * HWE p-values to <code>this.hweOut</code>.  The first output column
     * is marker id, followed by HWE p-values for each cohort (in the
     * order specified on the command line), followed by the HWE p-value
     * for the study (when all cohorts are combined).
     * 
     * @param cd input data for a marker.
     *
     * @return <code>true</code> if the marker data is excluded by HWE filter
     * and <code>false</code> otherwise, and writes HWE p-values to
     * <code>this.hweOut</code>.
     *
     * Reference: JE Wigginton, DJ Cutler, GR Abecasis (2005) A Note on
     * exact tests of Hardy-Weinberg equilibrium.  Am J Hum Genet 76:887-893.
     */
    private boolean hweFilter(CallingData cd) {
        double hwecohort = parameters.hwecohort();
        double hwestudy = parameters.hwestudy();
        double threshold = 0.0;
        String id = cd.markerId();
        boolean exclude = false;
        hweOut.print(id);
        if (cd.nGenotypes()==3) {
            int[] cumCnt = new int[cd.nGenotypes()];
            for (int j=0, n=cd.nCohorts(); j<n; ++j) {
                int[] cnt = getCounts(cd.gprobs(j), threshold);
                assert (cnt.length == cumCnt.length);
                for (int k=0; k<cumCnt.length; ++k) {
                    cumCnt[k] += cnt[k];
                }
                double pval = SNP_HWE.hweExact(cnt[0], cnt[1], cnt[2]);
                hweOut.print(tab);
                hweOut.print(dfsig2.format(pval));
                if (pval < hwecohort) {
                    exclude = true;
                }
            }
            double pval = SNP_HWE.hweExact(cumCnt[0], cumCnt[1], cumCnt[2]);
            hweOut.print(tab);
            hweOut.println(dfsig2.format(pval));
            if (pval < hwestudy) {
                exclude = true;
            }
        }
        return exclude;
    }

    public static int[] getCounts(GprobsData gprobs, double threshold) {
        int nGenotypes = gprobs.nGenotypes();
        int nSamples = gprobs.nSamples();
        int n = nSamples*nGenotypes;
        int[] counts = new int[nGenotypes];
        for (int j=0; j<n; j+=nGenotypes) {
            int maxIndex = 0;
            for (int k=1; k<nGenotypes; ++k) {
                if (gprobs.normData(j+k) > gprobs.normData(j + maxIndex)) {
                    maxIndex = k;
                }
            }
            double max = gprobs.normData(j + maxIndex);
            if (max >= threshold) {
                if (max > 0.5) {
                    ++counts[maxIndex];
                }
                else {  // next block checks that there is a unique maximum
                    int maxCnt = 0;
                    for (int k=0; k<nGenotypes; ++k) {
                        if (gprobs.normData(j+k)==max) {
                            ++maxCnt;
                        }
                    }
                    if (maxCnt == 1) {
                        ++counts[maxIndex];
                    }
                }
            }
        }
        return counts;
    }

    private boolean missingFilter(CallingData cd) {
        double callThreshold = parameters.callthreshold();
        double missingCohort = parameters.missingcohort();
        double missingStudy = parameters.missingstudy();
        boolean exclude = false;
        String id = cd.markerId();
        int nGenotypes = cd.nGenotypes();
        missingOut.print(id);
        int cumMissingCnt = 0;
        int cumCnt = 0;
        for (int i=0, n=cd.nCohorts(); i<n; ++i) {
            GprobsData data = cd.gprobs(i);
            int length = data.nSamples() * data.nGenotypes();
            int missingCnt = 0;
            for (int j=0; j<length; j+=nGenotypes) {
                boolean isMissing = true;
                for (int k=0; k<nGenotypes; ++k) {
                    if (data.normData(j+k) >= callThreshold) {
                        isMissing = false;
                    }
                }
                if (isMissing) {
                    ++missingCnt;
                }
            }
            int cnt = data.nSamples();
            cumMissingCnt += missingCnt;
            cumCnt += cnt;
            double missing = missingCnt / (double) cnt;
            missingOut.print(tab);
            missingOut.print(df4.format(missing));
            if (missing > missingCohort) {
                exclude = true;
            }
        }
        double missing = cumMissingCnt / (double) cumCnt;
        if (missing > missingStudy) {
            exclude = true;
        }
        missingOut.print(tab);
        missingOut.println(df4.format(missing));
        return exclude;
    }

    private double[] genotypeFrequencies(CallingData cd) {
        double epsilon = 1.0e-6;   // min frequency to avoid division by zero
        int nGenotypes = cd.nGenotypes();
        double[] freq = new double[nGenotypes];
        for (int i=0, n=cd.nCohorts(); i<n; ++i) {
            GprobsData data = cd.gprobs(i);
            int length = data.nSamples() * data.nGenotypes();
            for (int j=0; j<length; j+=nGenotypes) {
                for (int k=0; k<nGenotypes; ++k) {
                    freq[k] += data.normData(j+k);
                }
            }
        }
        double sum = 0.0;
        for (int j=0; j<freq.length; ++j) {
            if (freq[j] == 0.0) {
                freq[j] = epsilon;
            }
            sum += freq[j];
        }
        for (int j=0; j<freq.length; ++j) {
            freq[j] /= sum;
        }
        return freq;
    }

    private PrintWriter[] createPrintWriters(String prefix, File[] files,
            String suffix) {
        PrintWriter[] out = new PrintWriter[files.length];
        for (int j=0; j<files.length; ++j) {
            File f = new File(prefix + "." + files[j].getName() + suffix);
            out[j] = isCompressed ? Utilities.createGzipPrintWriter(f)
                    : Utilities.createPrintWriter(f);
        }
        return out;
    }

    private String exclusionSummary() {
        StringBuilder sb = new StringBuilder(500);
        sb.append(nl);
        sb.append("Marker summary");
        sb.append(nl);
        sb.append("------------------------------------------");
        sb.append(nl);
        sb.append(nl);
        sb.append("Markers in first .signals file:  ");
        sb.append(nSignalsMarkers);
        sb.append(nl);
        sb.append("Markers in first .gprobs file:   ");
        sb.append(nGprobsMarkers);
        sb.append(nl);
        sb.append(nl);
        sb.append("Specified exclusions:            ");
        sb.append(priorExclusionCnt);
        sb.append(nl);
        sb.append("QC filter exclusions:            ");
        sb.append(runtimeExclusionCnt);
        sb.append(nl);
        sb.append("Analyzed markers:                ");
        sb.append((nGprobsMarkers - priorExclusionCnt - runtimeExclusionCnt));
        sb.append(nl + nl);
        sb.append("QC Filter Summary");
        sb.append(nl);
        sb.append("------------------------------------------");
        sb.append(nl + nl);
        sb.append("Markers failing HWE filter:      ");
        sb.append(hweExcludeCnt);
        sb.append(nl);
        sb.append("Markers failing missing filter:  ");
        sb.append(missExcludeCnt);
        sb.append(nl);
        sb.append("Markers failing both QC filters: ");
        sb.append((hweExcludeCnt + missExcludeCnt - runtimeExclusionCnt));
        sb.append(nl);
        return sb.toString();
    }

    private void printSummary(long startMillis) {
        long currentMillis = System.currentTimeMillis();
        String elapsedTime = Utilities.elapsedTime(currentMillis - startMillis);
        Utilities.duoPrint(logOut, exclusionSummary());
        Utilities.duoPrint(logOut, nl + "Elapsed Time: " + elapsedTime);
        Utilities.duoPrint(logOut, nl + version + " finished" + nl);
        Utilities.duoPrint(logOut, 
        "==============================================================="
                + nl);
    }

    private void closeResources() {
        dataIt.close();
        logOut.close();
        hweOut.close();
        missingOut.close();
        excludeOut.close();
        markersOut.close();
        for (PrintWriter pw : modelOut) {
            pw.close();
        }
        for (PrintWriter pw : gprobsOut) {
            pw.close();
        }
        for (PrintWriter pw : likeOut) {
            pw.close();
        }
    }

    private String[] beagleArgs() {
        String prefix = parameters.out();
        File[] sigFiles = parameters.signals();
        List<String> args = new ArrayList<String>(20);
        args.add("markers=" + prefix + ".markers");
        for (File f : sigFiles) {
            args.add("like=" + prefix + "." + f.getName() + ".like.gz");
        }
        if (parameters.excludesamples() != null) {
            args.add("excludecolumns=" + parameters.excludesamples());
        }
        args.add("verbose=true");
        args.add("lowmem=true");
        args.add("out=" + parameters.out());
        args.add("omitprefix=true");
        return args.toArray(new String[0]);
    }
}
