/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import beagleutil.GprobsData;
import java.util.Arrays;

/**
 * Class CallingData represents current genotype probabilities and
 * alleles signal intensity data for a marker for multiple cohorts.
 *
 * @author Brian L Browning
 */
public class CallingData {

    private final String markerId;
    private final int nAlleles;
    private final int nGenotypes;
    private final String[] alleles;
    private final GprobsData[] gprobs;
    private final SignalsData[] signals;

    /**
     * Constructs a CallingData object from the genotype probability data,
     * signal intensity data.  The sample identifiers in in the specified
     * <code>SignalsData</code> and <code>GprobsData</code> instances are
     * not checked for consistency.
     *
     * @param sigData an array of <code>SignalsData</code> for a marker.
     * @param gpData an array of <code>GprobsData</code> for a marker.
     *
     * @throws NullPointerexception <code>sigData==null || gpData==null</code>
     * or if any element of the specified arrays is <code>null</code>.
     * @throws IllegalArgumentException if
     * <code>sigData.length != gpData.length || sigData.length == 0</code>
     * or if the marker identifier, number of alleles,
     * and alleles identifiers are not consistent for all of the
     * <code>SignalsData</code> and <code>GprobsData</code> instances.
     */
    public CallingData(SignalsData[] sigData, GprobsData[] gpData) {
        checkLengths(sigData, gpData);
        checkIDandAlleles(gpData, sigData);
        this.markerId = gpData[0].markerId();
        this.nAlleles = gpData[0].nAlleles();
        this.nGenotypes = gpData[0].nGenotypes();
        this.alleles = gpData[0].alleles();
        this.signals = new SignalsData[sigData.length];
        this.gprobs = new GprobsData[gpData.length];
        for (int j=0; j<signals.length; ++j) {
            signals[j] = sigData[j];
            gprobs[j] = gpData[j];
        }
    }

    private void checkLengths(SignalsData[] sigData, GprobsData[] gpData) {
        if (sigData.length != gpData.length) {
            String s = "cohorts with signals data: " + sigData.length +
                    " != cohorts with gprobs data: " + gpData.length;
            throw new IllegalArgumentException(s);
        }
        if (sigData.length==0) {
            String s = "sigData.length==0";
            throw new IllegalArgumentException(s);
        }
    }

    private void checkIDandAlleles(GprobsData[] gpData, SignalsData[] sigData) {
        String id = gpData[0].markerId();
        int alleleCnt = gpData[0].nAlleles();
        int genotypeCnt = gpData[0].nGenotypes();
        String[] alleleIds = gpData[0].alleles();
        String msg = null;
        for (GprobsData d : gpData) {
            if (!d.markerId().equals(id)) {
                msg = "marker identifier conflict: " + id + " != " + d.markerId();
            }
            if (d.nAlleles()!= alleleCnt) {
                msg = "allele count conflict: " + alleleCnt + " != "
                        + d.nAlleles();
            }
            if (d.nGenotypes()!= genotypeCnt) {
                msg = "genotype count conflict: " + genotypeCnt + " != " +
                        + d.nGenotypes();
            }
            if (!Arrays.equals(d.alleles(), alleleIds)) {
                msg = "allele identifier conflict: " + Arrays.toString(alleleIds)
                        + " != " + Arrays.toString(d.alleles());
            }
        }
        for (SignalsData s : sigData) {
            if (!s.markerId().equals(id)) {
                msg = "marker identifier conflict: " + id + " != " + s.markerId();
            }
            if (s.nAlleles()!= alleleCnt) {
                msg = "allele count conflict: " + alleleCnt + " != "
                        + s.nAlleles();
            }
            if (!Arrays.equals(s.alleles(), alleleIds)) {
                msg = "allele identifier conflict: " + Arrays.toString(alleleIds)
                        + " != " + Arrays.toString(s.alleles());
            }
        }
        if (msg!=null) {
            throw new IllegalArgumentException(msg);
        }
    }

    /**
     * Returns the marker identifier.
     * @return the marker identifier.
     */
    public String markerId() {
        return markerId;
    }

    /**
     * Returns the number of alleles identifiers for the marker.
     * @return the number of alleles identifiers for the marker.
     */
    public int nAlleles() {
        return nAlleles;
    }

    /**
     * Returns the number of genotype identifiers for the marker
     * @return the number of genotype identifiers for the marker.
     */
    public int nGenotypes() {
        return nGenotypes;
    }

    /**
     * Returns the number of cohorts.
     * @return the number of cohorts.
     */
    public int nCohorts() {
        return signals.length;
    }
    

    /**
     * Returns the specified alleles identifier.
     * @param k an index of an alleles.
     * @return the specified alleles identifier.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nAlleles()</code>.
     */
    public String alleles(int k) {
        return alleles[k];
    }

    /**
     * Returns the alleles signal intensity data for the specified cohort.
     *
     * @param k an index of a cohort.
     * @return the alleles signal intensity data for the specified cohort.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nCohorts()</code>.
     */
    public SignalsData signals(int k) {
        return signals[k];
    }

    /**
     * Returns the genotype probability data for the specified cohort.
     *
     * @param k an index of a cohort.
     * @return the genotype probability data for the specified cohort.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nCohorts()</code>.
     */
    public GprobsData gprobs(int k) {
        return gprobs[k];
    }

    /**
     * Returns a string description of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to change.
     * @return a string description of <code>this</code>.
     */
    @Override
    public String toString() {
        String space = " ";
        String nl = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        sb.append(markerId);
        sb.append(space);
        sb.append(nAlleles);
        for (int j=0; j<nAlleles; ++j) {
            sb.append(space);
            sb.append(alleles(j));
        }
        sb.append(space);
        sb.append(nGenotypes);
        sb.append(nl);
        for (int j=0; j<signals.length; ++j) {
            sb.append(j);
            sb.append(space);
            Arrays.toString(signals);
            sb.append(nl);
            sb.append(j);
            sb.append(space);
            Arrays.toString(gprobs);
            sb.append(nl);
        }
        return sb.toString();
    }
}
