/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import beagleutil.GprobsData;
import beagleutil.GprobsIterator;
import java.io.*;
import java.util.*;
import blbutil.*;

/**
 * Class <code>CallingDataIterator</code> iterates through input
 * data for genotype calling and returns a <code>CallingData</code> instance 
 * for each marker in the specified .gprobs files.
 *
 * The specified .gprobs files must each contain the same markers in the
 * same order (after applying specified marker exclusions).  The .signals files
 * must contain all the markers in the .gprobs files and the markers must be
 * in the same order as in the .gprobs files; however, the .signals files can
 * also contain additional markers.
 * 
 * @author Brian L Browning
 */
public class CallingDataIterator implements FileIterator<CallingData> {

    private final String newLine = System.getProperty("line.separator");
    private final boolean isMultiAllelic;
    private final File[] signalsFiles;
    private final File[] gprobsFiles;
    private final SignalsIterator[] signalsIts;
    private final GprobsIterator[] gprobsIts;

    private final SignalsData[] signalsData;
    private final GprobsData[] gprobsData;
    private CallingData next;

    /**
     * Constructs a <code>CallingDataIterator</code>.  The sample identifiers
     * in corresponding elements of <code>signalsFiles</code> and
     * <code>gprobsFiles</code> are checked for consistency.
     *
     * @param signalsFiles an array of files in BEAGLE .signals file format.
     * @param gprobsFiles an array of files in BEAGLE .gprobs file format.
     * @param sampleExclusions sample identifiers to be excluded.
     * @param markerExclusions marker identifiers to be excluded.
     * @param isMultiAllelic returns <code>true</code> if the input data
     * files contain a field specifying the number of alleles and
     * <code>false</code> otherwise.  If <code>isMultiAllelic==false</code>
     * the marker is assumed to be diallelic.
     *
     * @throws NullPointerException if
     * <code>signalsFiles==null || gprobsFiles==null 
     * || sampleExclusions==null || markerExclusions==null</code>.
     *
     * @throws IllegalArgumentException if
     * <code>signalsFiles.length != gprobsFiles.length</code> or if any files
     * are improperly formatted or if there is an inconsistency in the
     * samples in <code>signalsFiles[k]</code> and <code>gprobsFiles[k]</code>
     * for <code>0 &le; k < signalsFiles.length</code>.
     */
    public CallingDataIterator(File[] signalsFiles, File[] gprobsFiles,
            ImmutableSet<String> sampleExclusions,
            ImmutableSet<String> markerExclusions, boolean isMultiAllelic) {
        checkParameters(signalsFiles, gprobsFiles);
        this.isMultiAllelic = isMultiAllelic;
        this.signalsFiles = signalsFiles.clone();
        this.gprobsFiles = gprobsFiles.clone();
        this.signalsIts = new SignalsIterator[signalsFiles.length];
        this.gprobsIts = new GprobsIterator[gprobsFiles.length];
        this.signalsData = new SignalsData[signalsFiles.length];
        this.gprobsData = new GprobsData[gprobsFiles.length];
        for (int j=0; j<this.signalsFiles.length; ++j) {
            this.signalsIts[j] = new SignalsIterator(this.signalsFiles[j],
                    sampleExclusions, markerExclusions, isMultiAllelic);
            this.gprobsIts[j] = new GprobsIterator(this.gprobsFiles[j],
                    sampleExclusions, markerExclusions, isMultiAllelic);
        }
        this.next = readData();
        checkNextSamples();     // needs done only for first marker
    }

    private void checkParameters(File[] signalsFiles, File[] gprobsFiles) {
        if (signalsFiles.length != gprobsFiles.length) {
            String s = "signalsFiles.length=" + signalsFiles.length +
                    " != gprobsFiles.length=" + gprobsFiles.length;
            throw new IllegalArgumentException(s);
        }
        if (signalsFiles.length==0) {
            String s = "signalsFiles.length=0";
            throw new IllegalArgumentException(s);
        }
    }

    private void checkNextSamples() {
        for (int j=0, n=next.nCohorts(); j<n; ++j) {
            String[] signalsSamples = next.signals(j).sampleIds();
            String[] gprobsSamples = next.gprobs(j).sampleIds();
            if (!Arrays.equals(signalsSamples, gprobsSamples)) {
                String s = "samples in .signals file and .gprobs files "
                        + "do not match"
                        + newLine + signalsFiles[j].getName() + ": "
                        + Arrays.toString(signalsSamples) + newLine
                        + gprobsFiles[j].getName() + ": "
                        + Arrays.toString(gprobsSamples);
                throw new IllegalArgumentException(s);
            }
        }
    }

    public void close() {
        for (int j=0; j<this.signalsIts.length; ++j) {
            signalsIts[j].close();
            gprobsIts[j].close();
        }
        next = null;
    }

    public boolean hasNext() {
        return (next != null);
    }

    public CallingData next() {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        CallingData current = next;
        next = readData();
        return current;
    }

    public void remove() {
        String s = "remove() is not supported by CallingDataIterator";
        throw new UnsupportedOperationException(s);
    }

    private CallingData readData() {
        String id = null;
        for (int j=0; j<this.gprobsIts.length; ++j) {
            gprobsData[j] = null;
            signalsData[j] = null;
            if (gprobsIts[j].hasNext()) {
                gprobsData[j] = gprobsIts[j].next();
            }
            if (gprobsData[j] != null) {
                id = gprobsData[j].markerId();
                while (signalsIts[j].hasNext() && signalsData[j]==null) {
                    SignalsData candidate = signalsIts[j].next();
                    if (id.equals(candidate.markerId())) {
                        signalsData[j] = candidate;
                    }
                }
            }
        }
        int nGprobsNull = Utilities.countNull(gprobsData);
        int nSignalsNull = Utilities.countNull(signalsData);
        if (nGprobsNull==0 && nSignalsNull==0) {
            return new CallingData(signalsData, gprobsData);
        }
        else if (nGprobsNull==gprobsData.length
                && nSignalsNull==signalsData.length) {
            return null;
        }
        else {
            String s = "Marker " + id + " missing or out of order in one " +
                    "or more input files.";
            throw new IllegalArgumentException(s);
        }
    }

    /**
     * Returns a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to change.
     * @return a string representation of <code>this</code>.
     */
    @Override
    public String toString() {
        String space = " ";
        StringBuilder sb = new StringBuilder();
        sb.append(".signals files: ");
        sb.append(Arrays.toString(signalsFiles));
        sb.append(space);
        sb.append(".gprobs files: ");
        sb.append(Arrays.toString(gprobsFiles));
        sb.append(space);
        sb.append("isMultialllelic=");
        sb.append(isMultiAllelic);
        sb.append(space);
        sb.append(" next=");
        sb.append(next);
        return sb.toString();
    }
}
