/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import java.io.File;
import java.util.*;
import blbutil.Utilities;
import blbutil.Validate;

/**
 * Class CallingParameters represents the command line parameters and default
 * parameters for the BEAGLECALL program.
 * 
 * @author Brian L Browning
 */
public final class CallingParameters {

    public static final String newLine = System.getProperty("line.separator");
    private static final String usage = setUsage();

    private final File[] signals;
    private final File[] gprobs;
    private final File excludesamples;
    private final File excludemarkers;
    private final double callthreshold;
    private final double missingcohort;
    private final double missingstudy;
    private final double hwecohort;
    private final double hwestudy;
    private final double df;
    private final double assaysuccess;
    private final String out;
    private final int maxit;
    private final boolean runbeagle;
    private final double maxlr;

    /**
     * Constructs a new instance of <code>CallingParameters</code> for the
     * specified command line.
     *
     * @param args the command lilne arguments.
     * @throws IllegalArgumentException if there is an error in the command
     * line.
     */
    public CallingParameters(String[] args) {
        List<String> argsList = Utilities.argsToList(args, '=');
        if (argsList.size() > 0) {
            System.out.println("unrecognized command line parameters: "
                    + argsList);
        }
        Map<String, String[]> argsMap = Utilities.argsToMultimap(args, '=');

        this.signals = Validate.getFiles(
                Validate.mString("signals", argsMap, true, null));
        this.gprobs = Validate.getFiles(
                Validate.mString("gprobs", argsMap, true, null));
        this.out = Validate.sString("out", argsMap, true, null, null);
        this.callthreshold = Validate.sDouble("callthreshold", argsMap,
                false, 0.0, 0.0, 1.0);
        this.missingcohort = Validate.sDouble("missingcohort", argsMap, false, 1.0,
                0.0, 1.0);
        this.missingstudy = Validate.sDouble("missingstudy", argsMap, false, 1.0,
                0.0, 1.0);
        this.hwecohort = Validate.sDouble("hwecohort", argsMap, false, 1.0e-06,
                0.0, 1.0);
        this.hwestudy = Validate.sDouble("hwestudy", argsMap, false, 1.0e-06,
                0.0, 1.0);
        this.excludemarkers = Validate.getFile(Validate.sString("excludemarkers",
                argsMap, false, null, null));
        this.excludesamples = Validate.getFile(Validate.sString("excludesamples",
                argsMap, false, null, null));
        this.df = Validate.sDouble("df", argsMap, false, 5.0, 
                Utilities.nextUp(2.0), Double.MAX_VALUE);
        this.maxit = Validate.sInt("maxit", argsMap, false, 50, 1, Integer.MAX_VALUE);
        this.assaysuccess = Validate.sDouble("assaysuccess", argsMap, false,
                0.997, 0.0, 1.0);
        this.runbeagle = Validate.sBoolean("runbeagle", argsMap, false, true);
        this.maxlr = Validate.sDouble("maxlr", argsMap, false, 5000.0, 1.1,
                Double.MAX_VALUE);
        Validate.confirmEmptyMap(argsMap);
    }

    /**
     * Returns the command line help.
     *
     * @return the command line help.
     */
    public static String usage() {
        return usage;
    }

    private static String setUsage() {
        return  "Command line syntax: java -jar beaglecall.jar [arguments]" + newLine
                + newLine
                + "where [arguments] are ..." + newLine
                + "  signals=<allele signals file>                     (required)" + newLine
                + "  gprobs=<genotype probabilities file>              (required)" + newLine
                + "  out=<output file prefix>                          (required)" + newLine
                + "  callthreshold=<min genotype probability>          (optional: default=0.0)" + newLine
                + "  missingcohort=<max cohort missing proportion>     (optional: default=1.0)" + newLine
                + "  missingstudy=<max study missing proportion>       (optional: default=1.0)" + newLine
                + "  hwecohort=<min cohort HWE p-value>                (optional: default=1.0e-06" + newLine
                + "  hwestudy=<min study HWE p-value>                  (optional: default=1.0e-06" + newLine
                + "  excludemarkers=<excluded marker file>             (optional)" + newLine
                + "  excludesamples=<excluded samples file>            (optional)" + newLine
                + "  df=<degrees of freedom>                           (optional: default=5.0)" + newLine
                + "  maxit=<max iterations>                            (optional: default=50)" + newLine
                + "  assaysuccess=<initial assay success probability>  (optional: default=0.997)" + newLine
                + "  runbeagle=<true/false>                            (optional: default=true)" + newLine
                + "  maxlr=<max likelihood ratio>                      (optional: default=5000.0)" + newLine;
    }

    /**
     * Returns the signals file (set via command line).
     *
     * @return the signals file (set via command line).
     */
    public File[] signals() {
        return signals;
    }

    /**
     * Returns the prior genotype probability file (set via command line).
     *
     * @return the prior genotype probability file (set via command line).
     */
    public File[] gprobs() {
        return gprobs;
    }


    /**
     * Returns the output file prefix (set via command line).
     *
     * @return the output file prefix (set via command line).
     */
    public String out() {
        return out;
    }

    /**
     * Returns the calling threshold parameter (set via command line).  The
     * calling threshold is a threshold on the maximum genotype 
     * probability required to classify the genotype as
     * non-missing.
     *
     * @return the calling threshold parameter (set via command line).
     */
    public double callthreshold() {
        return callthreshold;
    }

    /**
     * Returns the maximum permitted missing proportion for a marker for
     * a cohort.  This parameter is set via the command line.
     *
     * @return the maximum permitted missing proportion for a marker for
     * a cohort.
     */
    public double missingcohort() {
        return missingcohort;
    }

    /**
     * Returns the maximum permitted missing proportion for a marker for
     * the study.  This parameter is set via the command line.
     *
     * @return the maximum permitted missing proportion for a marker 
     * for the study.
     */
    public double missingstudy() {
        return missingstudy;
    }

    /**
     * Returns the minimum permitted cohort HWE P-value parameter (set via
     * command line).
     *
     * @return the minimum permitted cohort HWE P-value parameter
     */
    public double hwecohort() {
        return hwecohort;
    }

    /**
     * Returns the minimum permitted study HWE P-value parameter (set via
     * command line).
     *
     * @return the minimum permitted study HWE P-value parameter
     */
    public double hwestudy() {
        return hwestudy;
    }

    /**
     * Returns the excluded markers file (set via command line) or
     * <code>null</code> if no excluded markers file was specified.
     *
     * @return the excluded markers file (set via command line) or
     * <code>null</code> if no excluded markers file was specified.
     */
    public File excludemarkers() {
        return excludemarkers;
    }

    /**
     * Returns the excluded samples file (set via command line) or
     * <code>null</code> if no excluded samples file was specified.
     *
     * @return the excluded samples file (set via command line) or
     * <code>null</code> if no excludedsamples file was specified.
     */
    public File excludesamples() {
        return excludesamples;
    }

    /**
     * Returns the degrees of freedom parameter (set via command line).
     *
     * @return the degrees of freedom parameter (set via command line).
     */
    public double df() {
        return df;
    }

    /**
     * Returns the maximum number of iterations that will be performed when
     * estimating the allele signal intensity model parameters. The
     * <code>maxit</code> parameter is set via the command line.
     *
     * @return  the maximum number of iterations that will be performed when
     * estimating the allele signal inensity model parameters.
     */
    public int maxit() {
        return maxit;
    }

    /**
     * Returns the initial assay assaysuccess probability parameter (set via
     * command line).
     *
     * @return the initial assay assaysuccess probability parameter (set via
     * command line).
     */
    public double assaysuccess() {
        return assaysuccess;
    }

    /**
     * Returns <code>true</code> if the posterior genotype probabilities 
     * should be generated using the BEAGLE haplotype frequency model, and
     * <code>false</code> if linkage disequilibrium should not be used
     * to generate posterior genotype probabilities. The
     * <code>runbeagle</code> parameter is set via the command line.
     *
     * @return <code>true</code> if the posterior genotype probabilities
     * should be generated using the BEAGLE haplotype frequency model, and
     * <code>false</code> if linkage disequilibrium should not be used
     * to generate posterior genotype probabilities.  The
     * <code>runbeagle</code> parameter is set via the command line.
     */
    public boolean runbeagle() {
        return runbeagle;
    }

    /**
     * Returns the maximum likelihood ratio parameter (set via the command
     * line).
     * @return the maximum likelihood ratio parameter (set via the command
     * line).
     */
    public double maxlr() {
        return maxlr;
    }
}
