/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */package calling;

import beagleutil.GprobsData;
import java.text.DecimalFormat;
import java.util.*;

/**
 * Class for modeling normalized allele signal data.
 *
 * @author Brian L Browning
 */
public class ModelMarker {
    
    public static final String SPACE = "\t";
    public static final String[] GENOTYPE_CLASSES = new String[] {"AA", "AB", "BB"};
    public static final double MAX_DF = 10000.0;
    private static final double MISSING_LIKELIHOOD = 1.0/3.0;
    private static final double MIN_THRESHOLD = 0.34;   // for uninformative input genotype probabilities
    private static final DecimalFormat df4 = new DecimalFormat("0.0###");

    private final double WEIGHT_THRESHOLD = 5.0;
    private final double REL_CONV_THRESHOLD = 0.001;
    private final double MIN_ABS_MIN = 1.0e-50;
    
    private SignalsData signalsData;
    private GprobsData gprobsData;
    private double[] df;
    private final double initAssaySuccess;
    private int it;
    private int maxIts;

    private double[] rangeA;        // { min(rangeA), max(rangeA) }
    private double[] rangeB;        // { min(rangeB), max(rangeB) }
    private ProbabilityDensity[] p1;// density for successful assays
    private ProbabilityDensity p0;  // density for failed assay
    private double[] frequencies;   // population genotype frequencies
    private double[] successProb;       // current individual assay success probabilities
    private double meanSuccessProb;     // current population assay success probability
    private double lastMeanSuccessProb; // previous population assay success probability

    private final double[] postGprobs;  // posterior genotype probabilities
    private final double[] postLike; // posterior genotype likelihoods
    
    private Moments moments = null;      // current estimated moments
    private Moments lastMoments = null;  // previous estimated momements

    private double[] pt = new double[2];

    /**
     * Constructs a <code>ModelMarker</code> instance for the specified
     * signal data, individual genotype probabilities, population genotype
     * frequencies and initial estimate of the assay success probability.
     *
     * @param signalsData allele signal intensity data for a marker.
     * @param gprobsData currect genotype probability estimates for a marker.
     * @param parameters the command line parameters for the analysis.
     * @param genotypeFrequencies an array with length equal to the number of
     * genotype classes, and whose j-th element is the population genotype
     * frequency for the j-th genotype class.
     *
     * @throws IllegalArgumentException if the specified signals data and
     * genotype probability data is not consistent or if
     * <code>genotypeFrequencies.length != gprobsData.nGenotypes()</code>.
     *
     * @throws NullPointerException if
     * <code>signalsData==null || gprobsData==null || parameters==null
     * || genotypeFrequencies==null</code>.
     */
    public ModelMarker(SignalsData signalsData, GprobsData gprobsData,
            CallingParameters parameters, double[] genotypeFrequencies) {
        checkParameters(signalsData, gprobsData, parameters,
                genotypeFrequencies);
        this.signalsData = signalsData;
        this.gprobsData = gprobsData;
        int n = this.gprobsData.nGenotypes() * this.gprobsData.nSamples();
        this.postGprobs = new double[n];
        this.postLike = new double[n];

        this.df = new double[this.gprobsData.nGenotypes()];
        Arrays.fill(this.df, parameters.df());
        this.initAssaySuccess = parameters.assaysuccess();
        this.frequencies = genotypeFrequencies.clone();
        this.maxIts = parameters.maxit();

        this.successProb = new double[this.gprobsData.nSamples()];
        Arrays.fill(successProb, initAssaySuccess);
        this.meanSuccessProb = initAssaySuccess;
        this.p1 = new ProbabilityDensity[this.gprobsData.nGenotypes()];
        this.p0 = new Uniform(signalsRange(signalsData.normData()));
        double[][] range = signalsRange(signalsData.normData());
        this.rangeA = range[0];
        this.rangeB = range[1];
        estimateParameters();
    }

    private static void checkParameters(SignalsData signalsData, 
            GprobsData gprobsData, CallingParameters parameters,
            double[] genotypeFrequencies) {
        checkDataConsistency(signalsData, gprobsData);
        if (genotypeFrequencies.length != gprobsData.nGenotypes()) {
            String s = "gprobsData.nGenotypes()=" + gprobsData.nGenotypes() 
                    + " != genotypeFrequencies.length="
                    + genotypeFrequencies.length;
            throw new IllegalArgumentException(s);
        }
        assert parameters.df() >= 2.0;
        assert parameters.assaysuccess() >= 0.0;
        assert parameters.assaysuccess() <= 1.0;
        assert parameters.maxit() > 0;
    }

    private static void checkDataConsistency(SignalsData sig, GprobsData gprob) {
        String nl = CallingParameters.newLine;
        if (!sig.markerId().equals(gprob.markerId())) {
            String s = "signalsData.markerId()=" + sig.markerId()
                    + ", but gprobsData.markerId()=" + gprob.markerId();
            throw new IllegalArgumentException(s);
        }
        String[] signalsAlleles = sig.alleles();
        String[] gprobsAlleles = gprob.alleles();
        if (!Arrays.equals(signalsAlleles, gprobsAlleles)) {
             String s = "alleles in .signals file and .gprobs files "
                    + "do not match for marker " + gprob.markerId() + nl
                    + Arrays.toString(signalsAlleles) + nl
                    + Arrays.toString(gprobsAlleles);
            throw new IllegalArgumentException(s);
        }
        String[] signalsSamples = sig.sampleIds();
        String[] gprobsSamples = gprob.sampleIds();
        if (!Arrays.equals(signalsSamples, gprobsSamples)) {
            String s = "samples in .signals file and .gprobs files "
                    + "do not match" + nl
                    + Arrays.toString(signalsSamples) + nl
                    + Arrays.toString(gprobsSamples);
            throw new IllegalArgumentException(s);
        }
    }

    private double[][] signalsRange(double[] signals) {
        double[] a = new double[] {Double.MAX_VALUE, -Double.MAX_VALUE};
        double[] b = new double[] {Double.MAX_VALUE, -Double.MAX_VALUE};
        for (int j=0; j<signals.length; j+=2) {
            double d = signals[j];
            if (d < a[0]) {
                a[0] = d;
            }
            if (d > a[1]) {
                a[1] = d;
            }
            d = signals[j+1];
            if (d < b[0]) {
                b[0] = d;
            }
            if (d > b[1]) {
                b[1] = d;
            }
        }
        return new double[][] {a, b};
    }

    /**
     * Returns the allele signal intensity data used to estimate posterior
     * genotype probabilities and genotype likelihoods.
     * @return the allele signal intensity data used to estimate posterior
     * genotype probabilities and genotype likelihoods.
     */
    public SignalsData signalsData() {
        return signalsData;
    }

    /**
     * Returns the current estimates of genotype probabilities used to
     * estimate posterior genotype probabilities and genotype likelihoods.
     * @return the current estimates of genotype probabilities used to
     * estimate posterior genotype probabilities and genotype likelihoods.
     */
    public GprobsData gprobsData() {
        return gprobsData();
    }

    /**
     * Returns an array of length <code>this.gprobsData().nGenotypes()</code>
     * whose <code>j</code>-th element is the degrees of freedom of the
     * t-distribution of the allele signal intensity data when the true
     * genotype is the <code>j-th</code> genotype class and the genotype
     * assay is successful
     * @return the degrees of freedom for the t-distributions that model
     * the allele signal intensity data when the genotype assay is successful.
     */
    public double[] df() {
        return df.clone();
    }

    /**
     * Returns the maximum number of iterations used to estimate parameters
     * of the probability model for the allele signal intensity data.
     * @return the maximum number of iterations used to estimate parameters
     * of the probability model for the allele signal intensity data.
     */
    public int maxIterations() {
        return maxIterations();
    }

    /**
     * Returns a <code>GprobsData</code> instance encoding the posterior
     * genotype likelihoods.
     * @return a <code>GprobsData</code> instance encoding the posterior
     * genotype likelihoods.
     */
    public GprobsData postLike() {
        return new GprobsData(gprobsData, postLike);
    }

    /**
     * Returns a <code>GprobsData</code> instance encoding the posterior
     * genotype probabilities.
     * @return a <code>GprobsData</code> instance encoding the posterior
     * genotype probabilities.
     */
    public GprobsData postGprobs() {
        return new GprobsData(gprobsData, postGprobs);
    }

    /**
     * Returns the estimated population assay success probability.
     * @return the estimated population assay success probability.
     */
    public double meanSuccessProb() {
        return meanSuccessProb;
    }

    /**
     * Replace the current estimated genotype probabilities with the
     * specified genotype probabilities, and re-estimate posteror genotype
     * probabilities and genotype postLike.
     * @param gprobs genotype probability data that will replace
     * <code>this.gprobsData.data()</code>.
     *
     * @throws NullPointerException if <code>gprobs==null</code>.
     * @throws IllegalArgumentException if the specified genotype
     * probabilities are not consistent with <code>this.signalsData()</code>.
     */
    public void update(GprobsData gprobs) {
        this.gprobsData = new GprobsData(gprobsData, gprobs.data());
        reestimateParameters();
    }

    /**
     * Models the allele signal intensity data using the specified degrees of
     * freedom.
     * @param df an array of length <code>this.df().length</code> specifying
     * the degrees of freedom for each genotype class.
     * @throws IllegalArgumentException if
     * <code>df.length != this.df.length</code> or if
     * <code>df[j] <= 2.0</code> for any <code>0 &le j < df.length</code>.
     */
    public void update(double[] df) {
        if (this.df.length != df.length) {
            String s = "df.length=" + df.length + " != this.df.length="
                    + this.df.length;
            throw new IllegalArgumentException(s);
        }
        for (int j=0; j<df.length; ++j) {
            if (df[j] <= 2.0) {
                String s = "df[" + j + "] == " + df[j] + " <= 2.0";
                throw new IllegalArgumentException(s);
            }
        }
        this.df = df.clone();
        reestimateParameters();
    }

    private void reestimateParameters() {
        this.it = 0;
        Arrays.fill(this.successProb, initAssaySuccess);
        this.meanSuccessProb = initAssaySuccess;
        estimateParameters();
    }

    private void estimateParameters() {
        boolean finished = false;
        updateMomentsAndP1();
        while (it<maxIts && !finished) {
            updateAssaySuccessProbability();
            updateMomentsAndP1();
            finished = isFinished();
            ++it;
        }
        setOutput(this.postLike, this.postGprobs);
    }

    private ProbabilityDensity sparseClusterDistribution(int j) {
        if (this.gprobsData.nGenotypes() != 3) {
            return p0;
        }
        double[] A = rangeA.clone();
        double[] B = rangeB.clone();
        if (j==0 && moments.sumWeights(2) > 4.0) {
            // cluster AA is sparse: restrict max B coordinate to mean of BB cluster
            B[1] = moments.mean(2)[1];
            return new Uniform(new double[][] {A, B});
        }
        else if (j==2 && moments.sumWeights(0) > 4.0) {
            // cluster BB is sparse: restrict max A coordinate to mean of AA cluster
            A[1] = moments.mean(0)[0];
            return new Uniform(new double[][] {A, B});
        }
        else {
            return p0;
        }
    }


    private double distance(double[] a, double[] b) {
        if (a.length != b.length) {
            String s = "a.length=" + a.length + " != b.length=" + b.length;
            throw new IllegalArgumentException(s);
        }
        double d = 0.0;
        for (int j=0; j<a.length; ++j) {
            d += Math.pow(a[j] - b[j], 2.0);
        }
        return Math.sqrt(d);
    }

    private void updateMomentsAndP1() {
        double[] weights = gprobsData.normData();
        for (int j=0; j<successProb.length; ++j) {
            int index = 3*j;
            // ignore individuals with uninformative prior genotype probabilities
            if (weights[index] <= MIN_THRESHOLD
                    && weights[index + 1] <= MIN_THRESHOLD
                    && weights[index + 2] <= MIN_THRESHOLD) {
                weights[index] = weights[index+1] = weights[index+2] = 0.0;
            }
            weights[index] *= successProb[j];
            weights[index + 1] *= successProb[j];
            weights[index + 2] *= successProb[j];
        }
        this.lastMoments = this.moments;
        this.moments = new Moments(signalsData.normData(), weights);
        for (int j=0, n=this.gprobsData.nGenotypes(); j<n; ++j) {
            double[][] var = moments.variance(j);
            double det = var[0][0]*var[1][1] - var[0][1]*var[1][0];
            if (moments.sumWeights(j) < 1.0 || det < 1e-8) { // det may be < 0.0
                // reduce likelihoods by a factor of ~2*nSamples
                double c = gprobsData.nSamples()*(rangeA[1] - rangeA[0]);
                double[] a = new double[] {(rangeA[0]-c), (rangeA[1]+c)};
                p1[j] = new Uniform(new double[][] {a, rangeB});
            }
            else if (moments.sumWeights(j)<WEIGHT_THRESHOLD || det < 1e-8) {  // det may be < 0.0
                    p1[j] = sparseClusterDistribution(j);  // ideally should use training data to set p1
                }
            else {
                double[] m = moments.mean(j);
                double[][] v = moments.variance(j);
//                for (int r=0; r<v.length; ++r) {
//                    for (int s=0; s<v[r].length; ++s) {
//                        v[r][s] = ((df[j]-2.0)/df[j])*v[r][s];
//                    }
//                }
                p1[j] = (df[j]>MAX_DF) ? new Normal(m, v) : new T(m, v, df[j]);
            }
        }
    }

    /*
     * Updates the posterior probability of assay success for each
     * individual P(Z_i = 1 | S_i) and updates the mean assay success
     * probability
     */
    private void updateAssaySuccessProbability() {
        int nGenotypes = this.gprobsData.nGenotypes();
        lastMeanSuccessProb = meanSuccessProb;
        assert meanSuccessProb >= 0.0 && meanSuccessProb <= 1.0;

        double lastMeanFailureProb = 1.0 - lastMeanSuccessProb;
        double sum = 0.0;
        assert successProb.length == signalsData.nSamples();
        for (int j=0; j<successProb.length; ++j) {
            pt[0] = signalsData.normData(2*j);
            pt[1] = signalsData.normData(2*j + 1);
            if (Double.isNaN(pt[0]) || Double.isNaN(pt[1])) {
                this.successProb[j] = 0.0;
            }
            else {
                double p0Prob = p0.probability(pt);
                double a = 0.0;
                double b = p0Prob * lastMeanFailureProb;
                for (int k=0; k<nGenotypes; ++k) {
                    int index = j*nGenotypes + k;
                    a += p1[k].probability(pt) * gprobsData.normData(index)
                            * lastMeanSuccessProb;
                }
                this.successProb[j] = ((a+b)==0.0) ? 0.0 : a/(a+b);
            }
            sum += this.successProb[j];
            assert successProb[j] >= 0.0 && successProb[j] <= 1.0;
        }
        this.meanSuccessProb = sum/successProb.length;
    }

    private boolean isFinished() {
        int nGenotypes = this.gprobsData.nGenotypes();
        double maxAbsRelChange = 0;
        for (int j=0; j<nGenotypes; ++j) {
            if (moments.sumWeights(j) >= WEIGHT_THRESHOLD) {
                double[] oldMean = lastMoments.mean(j);
                double[] curMean = moments.mean(j);
                for (int k=0; k<oldMean.length; ++k) {
                    double d = maxAbsRelChange(oldMean[k], curMean[k]);
                    if (d > maxAbsRelChange) {
                        maxAbsRelChange = d;
                    }
                }
                double[][] oldVar = lastMoments.variance(j);
                double[][] curVar = moments.variance(j);
                for (int k=0; k<oldVar.length; ++k) {
                    for (int l=0; l<oldVar[k].length; ++l) {
                        double a = oldVar[k][l];
                        double b = curVar[k][l];
                        double change = maxAbsRelChange(a, b);
                        double absMin = Math.min(Math.abs(a), Math.abs(b));
                        if (change > maxAbsRelChange && absMin >= MIN_ABS_MIN) {
                             maxAbsRelChange = change;
                        }
                    }
                }
            }
        }
        double d = maxAbsRelChange(lastMeanSuccessProb, meanSuccessProb);
        if (d > maxAbsRelChange) {
            maxAbsRelChange = d;
        }
        return (maxAbsRelChange < REL_CONV_THRESHOLD);
    }

    private double maxAbsRelChange(double oldValue, double newValue) {
        if (oldValue==0.0) {
            return (newValue==0.0) ? 0.0 : 1.0;
        }
        else {
            return Math.abs((newValue - oldValue)/oldValue);
        }
    }

    private void setOutput(double[] postLikelihood, double[] postProbability) {
        int nGenotypeClasses = this.gprobsData.nGenotypes();
        int n = nGenotypeClasses * gprobsData.nSamples();
        assert postLikelihood.length == n;
        assert postLikelihood.length == postProbability.length;
        int sampleIndex = 0;
        int signalIndex = 0;
        for (int j=0; j<n; j+=nGenotypeClasses) {
            double probSum = 0.0;
            double likeSum = 0.0;
            pt[0] = signalsData.normData(signalIndex++);
            pt[1] = signalsData.normData(signalIndex++);
            if (Double.isNaN(pt[0]) || Double.isNaN(pt[1])) {
                for (int k=0; k<nGenotypeClasses; ++k) {
                    int index = j + k;
                    postProbability[index] = frequencies[k];
                    postLikelihood[index] = MISSING_LIKELIHOOD;
                }
            }
            else {
                double alt = p0.probability(pt)*(1-successProb[sampleIndex]);
                for (int k=0; k<nGenotypeClasses; ++k) {
                    int index = j + k;
                    double d = alt + p1[k].probability(pt)*successProb[sampleIndex];
                    postLikelihood[index] = d;
                    postProbability[index] = d*frequencies[k];
                    probSum += postProbability[index];
                    likeSum += d;
                }
                for (int k=0; k<nGenotypeClasses; ++k) {
                    int index = j + k;
                    postProbability[index] /= probSum;
                    postLikelihood[index] /= likeSum;
                }
            }
            ++sampleIndex;
        }
        assert sampleIndex==successProb.length;
        assert signalIndex==(signalsData.nSamples()*signalsData.nAlleles());
    }

    /**
     * Returns the maximum absolute change between input genotype
     * probabilities and the estimated posterior genotype probabilities.
     * @return the maximum absolute change between input genotype
     * probabilities and the estimated posterior genotype probabilities.
     */
    public double maxAbsoluteChange() {
        double maxChange = 0.0;
        int n = gprobsData.nSamples() * gprobsData.nGenotypes();
        for (int l=0; l<n; ++l) {
            double x = Math.abs(gprobsData.normData(l) - postGprobs[l]);
            if (x > maxChange) {
                maxChange = x;
            }
        }
        return maxChange;
    }

    /**
     * Returns a string description of the current model.  The format of the
     * description is unspecified and subject to change; however, the format
     * will typically include marker identifier, allele identifiers,
     * mean assay success probability, populaton genotype probabilities,
     * and a description of the probability density function (name, mean,
     * and variance).
     *
     * @return a string description fo the current model.
     */
    public String model() {
        if (gprobsData.nGenotypes() != 3) {
            return gprobsData.markerId() + " model not printed";
        }
        StringBuilder sb = new StringBuilder();
        sb.append(gprobsData.markerId());
        if (moments != null) {
            sb.append(SPACE);
            sb.append(it);
            sb.append(SPACE);
            sb.append(gprobsData.alleles(0));
            sb.append(SPACE);
            sb.append(gprobsData.alleles(1));
            sb.append(SPACE);
            sb.append(format(meanSuccessProb));
            for (int j=0; j<frequencies.length; ++j) {
                sb.append(SPACE);
                sb.append(format(frequencies[j]));
            }
            for (int j=0; j<GENOTYPE_CLASSES.length; ++j) {
                describeModel(j, sb);
            }
        }
        return sb.toString();
    }

    private void describeModel(int genotypeClass, StringBuilder sb) {
        if (p1[genotypeClass].name().equals("uniform")) {
            sb.append(SPACE);
            sb.append(GENOTYPE_CLASSES[genotypeClass]);
            sb.append(SPACE);
            sb.append("uniform");
            for (int k=0; k<5; ++k) {
                sb.append(SPACE);
                sb.append("---");
            }
        }
        else {
            double[] mean = moments.mean(genotypeClass);
            double[][] var = moments.variance(genotypeClass);
            sb.append(SPACE);
            sb.append(GENOTYPE_CLASSES[genotypeClass]);
            sb.append(SPACE);
            sb.append(format(df[genotypeClass]));
            sb.append(SPACE);
            sb.append(format(mean[0]));
            sb.append(SPACE);
            sb.append(format(mean[1]));
            sb.append(SPACE);
            sb.append(format(var[0][0]));
            sb.append(SPACE);
            sb.append(format(var[0][1]));
            sb.append(SPACE);
            sb.append(format(var[1][1]));
        }
    }

    private String format(double d) {
        if (Double.isNaN(d)) {
            return "NaN";
        }
        else {
            return df4.format(d);
        }
    }

    /**
     * Returns a string header line describing the columns in the string
     * returned by the <code>this.model()</code> method. The format of the
     * header line is unspecified and subject to change.
     * 
     * @return a string header line describing the columns in the string
     * returned by the <code>this.model()</code> method.
     */
    public static String modelHeaderLine() {
        String space = "\t";
        StringBuilder sb = new StringBuilder();
        sb.append("marker");
        sb.append(space);
        sb.append("iter");
        sb.append(space);
        sb.append("A");
        sb.append(space);
        sb.append("B");
        sb.append(space);
        sb.append("success");
        sb.append(space);
        sb.append("freq.AA");
        sb.append(space);
        sb.append("freq.AB");
        sb.append(space);
        sb.append("freq.BB");
        for (int j=0; j<2; ++j) {
            sb.append(space);
            sb.append("Gtype");
            sb.append(space);
            sb.append("df");
            sb.append(space);
            sb.append("mean.A");
            sb.append(space);
            sb.append("mean.B");
            sb.append(space);
            sb.append("var.AA");
            sb.append(space);
            sb.append("cov.AB");
            sb.append(space);
            sb.append("var.BB");
        }
        return sb.toString();
    }
}
