/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import java.util.Arrays;

/**
 * Class Moments has methods for returning the sample mean, sample variance,
 * and first and second sample moments of A and B signal intensities for
 * each genotype class.
 *
 * @author Brian L Browning
 */
public final class Moments {

    private final int n;
    private final int nGenotypes;
    private final double[] sumW;
    private final double[] sumA;
    private final double[] sumB;
    private final double[] sumAA;
    private final double[] sumAB;
    private final double[] sumBB;

    /**
     * Constructs a <code>Moments</code> object for the specified
     * signal intensities and weights.
     *
     * @param signals an array of length <code>2 * n</code> where
     * <code>n</code> is the number of individuals in the sample whose
     * <code>j * 2</code>-th element is the normalized signal intensity
     * for the A allele for the j-th individual and whose
     * <code>j * 2 + 1</code>-th element is the normalized signal intensity
     * for the B allele for the j-th individual.
     * @param weights  an array of length <code>nGenotypes * n</code> where
     * <code>nGenotypes</code> is the number of genotype classes, and
     * <code>n</code> is the number of individuals in the sample.
     * The <code>j * nGenotypes + k</code>-th element is the probability
     * that the j-th individual has genotype k and that the 
     * assay for the genotype was successful.
     */
    public Moments(double[] signals, double[] weights) {
        this.nGenotypes = checkLengths(signals, weights);
        this.n = signals.length / 2;
        this.sumW = new double[nGenotypes];
        this.sumA = new double[nGenotypes];
        this.sumB = new double[nGenotypes];
        this.sumAA = new double[nGenotypes];
        this.sumAB = new double[nGenotypes];
        this.sumBB = new double[nGenotypes];
        int signalIndex = 0;
        for (int j=0; j<weights.length; j+=nGenotypes) {
            double a = signals[signalIndex++];
            double b = signals[signalIndex++];
            if (!Double.isNaN(a) && !Double.isNaN(b)) {
                for (int k=0; k<nGenotypes; ++k) {
                    double w = weights[j + k];
                    sumW[k] += w;
                    sumA[k] += a*w;
                    sumB[k] += b*w;
                    sumAA[k] += a*a*w;
                    sumAB[k] += a*b*w;
                    sumBB[k] += b*b*w;
                }
            }
        }
        assert signalIndex == signals.length;
    }

    /*
     * Returns the number of genotype classes.
     */
    private int checkLengths(double[] signals, double[] weights) {
        if ((signals.length & 1)!= 0) {
            String s = "(signals.length % 2) != 0: " + signals.length;
            throw new IllegalArgumentException(s);
        }
        int size = signals.length / 2;
        if (size==0) {
            throw new IllegalArgumentException("signals.length==0");
        }
        if ((weights.length % size) != 0) {
            String s = "(weights.length % (signals.length/2)) != 0 :" +
                weights.length + " " + (signals.length/2);
            throw new IllegalArgumentException(s);
        }
        return weights.length / size;
    }

    /**
     * Returns a positive integer giving the number of genotype classes.
     * @return a positive integer giving the number of genotype classes.
     */
    public int nGenotypes() {
        return nGenotypes;
    }

    /**
     * Returns the sum of the weights for the specified genotype class. 
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return the sum of the weights for the specified genotype class. 
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double sumWeights(int genotype) {
        return sumW[genotype];
    }

    /**
     * Returns an array of length 2 giving an estimate of the mean of
     * the genotype cluster for the specified genotype class.
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return an array of length 2 giving an estimate of the mean of
     * the genotype cluster for the specified genotype class.
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double[] mean(int genotype) {
        double[] mean = new double[2];
        mean[0] = sumA[genotype] / sumW[genotype];
        mean[1] = sumB[genotype] / sumW[genotype];
        return mean;
    }

    /**
     * Returns a two-dimensional array with two rows and two columns
     * giving an estimate of the variance of
     * the genotype cluster for the specified genotype class.
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return a two-dimensional array with two rows and two columns
     * giving an estimate of the variance of
     * the genotype cluster for the specified genotype class.
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double[][] variance(int genotype) {
        double w = sumW[genotype];
        double[][] var = new double[2][2];
        double cov = (sumAB[genotype] - (sumA[genotype]*sumB[genotype])/w)/w;
        var[0][0] = (sumAA[genotype] - (sumA[genotype]*sumA[genotype])/w)/w;
        var[0][1] = cov;
        var[1][0] = cov;
        var[1][1] = (sumBB[genotype] - (sumB[genotype]*sumB[genotype])/w)/w;
        return var;
    }
    
    /**
     * Returns the number of samples.
     * @return the number of samples.
     */
    public int nSamples() {
        return n;
    }

    /**
     * Returns an estimate of the sum of the A allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return an estimate of the sum of the A allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double sumA(int genotype) {
        return sumA[genotype];
    }

    /**
     * Returns an estimate of the sum of the B allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return an estimate of the sum of the B allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double sumB(int genotype) {
        return sumB[genotype];
    }

    /**
     * Returns an estimate of the sum of the square of the A allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return an estimate of the sum of the square of the A allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double sumAA(int genotype) {
        return sumAA[genotype];
    }

    /**
     * Returns an estimate of the sum of the product of the A and B allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return an estimate of the sum of the product of the A and B allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double sumAB(int genotype) {
        return sumAB[genotype];
    }

    /**
     * Returns an estimate of the sum of the square of the B allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     * @param genotype a genotype class.  The genotype class must be a
     * nonnegative integer less than <code>this.nGenotypes()</code>.
     * @return an estimate of the sum of the square of the B allele
     * signal intensities for the genotype cluster for the specified
     * genotype class.
     *
     * @throws ArrayIndexOutOfBoundsException if
     * <code>genotype < 0 || genotype &ge; this.nGenotypes()</code>.
     */
    public double sumBB(int genotype) {
        return sumBB[genotype];
    }

    /**
     * Returns a string representation of <code>this</code>.  The
     * exact details of the representation are unspecified and
     * subject to change.
     * @return a string representation of <code>this</code>.
     */
    @Override
    public String toString() {
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder(500);
        sb.append("n=");
        sb.append(n);
        sb.append(" nGenotypes=");
        sb.append(nGenotypes);
        sb.append(" sumW=");
        sb.append(Arrays.toString(sumW));
        sb.append(newLine);
        sb.append("sumA=");
        sb.append(Arrays.toString(sumA));
        sb.append(" sumB=");
        sb.append(Arrays.toString(sumB));
        sb.append(newLine);
        sb.append("sumAA=");
        sb.append(Arrays.toString(sumAA));
        sb.append(newLine);
        sb.append("sumAB=");
        sb.append(Arrays.toString(sumAB));
        sb.append(newLine);
        sb.append("sumBB=");
        sb.append(Arrays.toString(sumBB));
        sb.append(newLine);
        return sb.toString();
    }
}
