/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import java.util.Arrays;
import org.apache.commons.math.linear.*;

/**
 * Probability density for a multivariate Normal distribution.
 * 
 * @author Brian L Browning
 */
public class Normal implements ProbabilityDensity {

    private final double[] mean;
    private final double c;
    private final RealMatrix variance;
    private final RealMatrix varInverse;

    /**
     * Constructs a <code>ProbabilityDensity</code> object
     * for a multivariate Normal distribution.
     * 
     * @param mean the distribution mean.
     * @param variance the distribution variance.
     *
     * @throws IllegalArgumentException if
     * <code>mean.length != variance.length</code> or if
     * any element of the mean and variance arrays are not finite numbers.
     * @throws NullPointerException if
     * <code>mean==null || variance==null || variance[j]==null</code>
     * for any <code>0 &le; j < variance.length</code>.
     */
    public Normal(double[] mean, double[][] variance) {
        checkDimensions(mean.length, variance.length);
        for (double[] da : variance) {
            checkDimensions(mean.length, da.length);
        }
        checkEntries(mean);
        for (double[] da : variance) {
            checkEntries(da);
        }
        this.mean = mean.clone();
        this.variance = new RealMatrixImpl(variance);
        this.varInverse = this.variance.inverse();
        double d = Math.pow(this.variance.getDeterminant(), -0.5);
        this.c = d*Math.pow((2*Math.PI), -mean.length/2.0);
    }

    private void checkDimensions(int d1, int d2) {
        if (d1!=d2) {
            String s = "dimensions are inconsistent: " + d1 + " " + d2;
            throw new IllegalArgumentException(s);
        }
    }

    private void checkEntries(double[] da) {
        for (double d : da) {
            if (Double.isNaN(d)) {
                String s = "ERROR: undefined parameter: " + Arrays.toString(da);
                throw new IllegalArgumentException(s);
            }
            if (Double.isInfinite(d)) {
                String s = "Error: infinite parameter: " + Arrays.toString(da);
                throw new IllegalArgumentException(s);
            }
        }
    }

    public double probability(double[] x) {
        if (x.length != mean.length) {
            String s = "x.length != dimension(): " + x.length + " " + dimension();
            throw new IllegalArgumentException(s);
        }
        double quadForm = 0;
        double[] shiftedX = new double[x.length];
        for (int j=0; j<mean.length; ++j) {
            if (Double.isNaN(x[j])) {
                return Double.NaN;
            }
            if (Double.isInfinite(x[j])) {
                return 0.0;
            }
            shiftedX[j] = x[j] - mean[j];
        }
        for (int i=0; i<mean.length; ++i) {
            for (int j=0; j<mean.length; ++j) {
                quadForm += shiftedX[i]*varInverse.getEntry(i, j)*shiftedX[j];
            }
        }
        return c*Math.exp(-(0.5)* quadForm);
    }

    public int dimension() {
        return mean.length;
    }

    public double[] mean() {
        return mean.clone();
    }

    public double[][] variance() {
        return variance.getData();
    }

    public double df() {
        return Double.NaN;
    }

    public String name() {
        return "normal";
    }

    /**
     * Returns a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to
     * change.
     * 
     * @return a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to
     * change.
     */
    @Override
    public String toString() {
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        sb.append("Normal Density");
        sb.append(newLine);
        sb.append("mean=" + Arrays.toString(mean));
        sb.append(newLine);
        sb.append("variance=" + variance);
        sb.append(newLine);
        sb.append("varInverse=" + varInverse);
        sb.append(newLine);
        return sb.toString();
    }

// <editor-fold defaultstate="collapsed" desc=" commented-out code for testing class ">
/* code for testing class */
//    public static void main(String[] args) {
//        double[] mean = new double[] {1, 2};
//        double[][] var = new double[2][];
//        var[0] = new double[] {8.0, 3.0};
//        var[1] = new double[] {2.0, 5.0};
//        ProbabilityDensity n = new Normal(mean, var);
//        System.out.println(n.probability(new double[] {Double.POSITIVE_INFINITY, Double.NaN}));
//        System.out.println(n.probability(new double[] {3.0, Double.NEGATIVE_INFINITY}));
//        System.out.println(n.probability(new double[] {3.0, Double.POSITIVE_INFINITY}));
//        int max = 10;
//        int inc = 4;
//        for (int i=-max; i<=max; i+=inc) {
//            for (int j=-max; j<=max; j+=inc) {
//                System.out.println("norm(" + i + ", " + j + ")=" + n.probability(new double[] {i, j}));
//            }
//        }
//    }
// </editor-fold>
}
