/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

/**
 * Interface for a multivariate probability distribution.
 *
 * @author Brian L Browning
 */
public interface ProbabilityDensity {

    /**
     * Returns the value of the probability density function at the specified
     * point.  Returns <code>Double.NaN</code> if any element of the
     * specified point is <code>Double.NaN</code>.  Returns 0.0
     * if any element of the specified point is infinite and no elements
     * of the specified point are <code>Double.NaN</code>.
     * @param x a vector of length <code>dimension()</code>.
     * @return the value of the probability distribution at the specified
     * point.
     *
     * @throws NullPointerException if <code>x==null</code>
     * @throws IllegalArgumentException if
     * <code>x.length != this.dimension()</code>.
     */
    public double probability(double[] x);

    /**
     * Returns a positive integer giving the dimension of the
     * probability distribution.
     * 
     * @return a positive integer giving the dimension of the
     * probability distribution.
     */
    public int dimension();

    /**
     * Returns a vector of length <code>this.dimension()</code> giving
     * the mean of the probability distribution.
     *
     * @return a vector of length <code>this.dimension()</code> giving
     * the mean of the probability distribution.
     */
    public double[] mean();

    /**
     * Returns a two-dimensional array with <code>this.dimension()</code>
     * rows and <code>this.dimension()</code> columns that parameterizes
     * the variance/covariance matrix for the probability distribution.
     *
     * Note that for the t-distribution, the variance of the t-distribution
     * is obtained by multiplying each element of <code>this.variance()</code>
     * by <code>(df()/(df() - 2.0)</code>.
     *
     * @return the variancec/covariance matrix parameters for the probability
     * distribution.
     */
    public double[][] variance();
    

    /**
     * Returns the degrees of freedom for the distribution or
     * <code>Double.NaN</code> if the degrees of freedom are not defined.
     *
     * @return the degrees of freedom for the distribution or
     * <code>Double.NaN</code> if the degrees of freedom are not defined.
     */
    public double df();

    /**
     * Returns a brief description of the probability densitiy function.
     * 
     * @return a brief description of the probability densitiy function.
     */
    public String name();
}
