/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import beagleutil.Samples;
import java.util.Arrays;
import blbutil.*;

/**
 * Class <code>SignalsData</code> represented a line of data
 * from a file in BEAGLE .signals file format.
 * 
 * @author Brian L Browning
 */
public class SignalsData {

    private final Samples samples;
    private final String markerId;
    private final int nAlleles;
    private final String[] alleles;
    private final double[] data;
    private final double[] normData;

    /**
     * Constructs a <code>SignalsData</code> instance from the specified
     * data line of a BEAGLE .signals file.
     *
     * @param line a non-header line of data from a file in BEAGLE .signals
     * format.
     * @param samples the filtered and unfiltered sample identifiers for the
     * data.
     * @param isMultiAllelic <code>true</code> if the specified
     * <code>line</code>  contains a field specifying the number of alleles
     * for the marker and <code>false</code> otherwise.
     * If <code>isMultiAllelic==true</code> the second field
     * of the line contains the number of alleles, and the third field
     * begins the enumeration of the marker alleles.
     * If <code>isMultiAllelic==false</code> the marker is assumed to be
     * a diallelic marker, and the second and third fields give the marker
     * alleles.
     *
     * @throws NullPointerException if <code>line==null || samples==null</code>.
     * @throws IllegalArgumentException if 
     * the number of alleles is less than 1, or if the
     * number of fields on the line is inconsistent with the number of
     * specified alleles and the number of unfiltered samples.
     * @throws NumberFormatException if any field that is expected to be an
     * integer or a double is not parsable.
     */
    public SignalsData(String line, Samples samples, boolean isMultiAllelic) {
        String[] fields = StringUtil.getFields(line);
        this.nAlleles = checkArguments(fields, samples, isMultiAllelic);
        this.samples = samples;
        this.markerId = fields[0];
        this.alleles = new String[this.nAlleles];
        int startIndex = isMultiAllelic ? 2 : 1;
        for (int j=0; j<this.alleles.length; ++j) {
            this.alleles[j] = fields[startIndex++];
        }
        this.data = new double[this.samples.filteredSize()*this.nAlleles];
        int index = 0;
        for (int j=0, n=samples.unfilteredSize(); j<n; ++j) {
            if (this.samples.filter(j)==false) {
                for (int k=0; k<this.nAlleles; ++k) {
                    double d = Double.parseDouble(
                            fields[startIndex + (j*nAlleles) + k]);
                    this.data[index+k] = d;
                }
                index += this.nAlleles;
            }
        }
        this.normData = this.data.clone();
    }

    /*
     * Returns the number of alleles for the marker.
     */
    private int checkArguments(String[] fields, Samples samples,
            boolean isMultiAllelic) {
        int alleleCnt = isMultiAllelic ? Integer.parseInt(fields[1]) : 2;
        if (alleleCnt < 1) {
            String s = "nAlleles < 1: " + alleleCnt + ": " +
                    Arrays.toString(fields);
            throw new IllegalArgumentException(s);
        }
        int expectedFields = 1 + alleleCnt + alleleCnt*samples.unfilteredSize();
        if (isMultiAllelic) {
            ++expectedFields;
        }
        if (fields.length != expectedFields) {
            String s = "Error for marker " + fields[0] + ": expected "
                    + expectedFields + " fields, found " + fields.length +
                    " fields (nSamples=" + samples.unfilteredSize() + 
                    " nAlleles=" + nAlleles + "): " + Arrays.toString(fields);
            throw new IllegalArgumentException(s);
        }
        return alleleCnt;
    }

    /**
     * Returns the marker identifier.
     * @return the marker identifier.
     */
    public String markerId() {
        return markerId;
    }

    /**
     * Returns the number of samples.
     * @return the number of samples.
     */
    public int nSamples() {
        return samples.filteredSize();
    }

    /**
     * Returns the specified sample identifier.
     * @param k an index of an sample.
     * @return the specified samples identifier.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.Samples()</code>.
     */
    public String sampleIds(int k) {
        return null;
    }


    /**
     * Returns an array of sample identifiers such that
     * <code>this.samples()[k].equals(this.sample(k))</code>
     * for <code> 0 &le; k < this.nSamples()</code>.
     *
     * @return an array of sample identifiers.
     */
    public String[] sampleIds() {
        return samples.filteredIds().clone();
    }

   /**
     * Returns the number of allele identifiers for the marker.
     * @return the number of allele identifiers for the marker.
     */
    public int nAlleles() {
        return nAlleles;
    }

    /**
     * Returns an array of allele identifiers.
     * The <code>this.alleles()[k].equals(this.alleles(k))</code>
     * for <code> 0 &le; k < this.nAlleles()</code>.
     *
     * @return an array of allele identifiers.
     */
    public String[] alleles() {
        return alleles.clone();
    }

    /**
     * Returns the specified allele identifier.
     * @param k an index of an allele.
     * @return the specified allele identifier.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nAlleles()</code>.
     */
    public String alleles(int k) {
        return alleles[k];
    }

    /**
     * Returns the array of allele signal intensity data for the marker.
     * @return the array of allele signal intensity data for the marker.
     */
    public double[] data() {
        return data.clone();
    }

    /**
     * Returns the normalized array of allele signal intensity data for the
     * marker.  The normalization method is unspecified.
     * @return the normalized array of allele signal intensity data for the
     * marker.
     */
    public double[] normData() {
        return normData.clone();
    }

    /**
     * Returns <code>data()[j]</code>.
     * @param j an array index.
     * @return <code>data()[j]</code>.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nSamples()*this.nAlleles()</code>.
     */
    public double data(int j) {
        return data[j];
    }

    /**
     * Returns <code>normData()[j]</code>.
     * @param j an array index.
     * @return <code>normData()[j]</code>.
     * @throws ArrayIndexOutOfBoundsException if
     * <code> k < 0 || k &ge; this.nSamples()*this.nAlleles()</code>.
     */
    public double normData(int j) {
        return normData[j];
    }

    @Override
    public String toString() {
        String space = " ";
        StringBuilder sb = new StringBuilder();
        sb.append(markerId);
        sb.append(space);
        sb.append(nAlleles);
        for (int j=0; j<nAlleles; ++j) {
            sb.append(space);
            sb.append(alleles(j));
        }
        sb.append(space);
        sb.append(samples.filteredSize());
        for (int j=0, n=samples.filteredSize(); j<n; ++j) {
            sb.append(space);
            sb.append(samples.filteredIds(j));
            sb.append("=");
            sb.append(Arrays.toString(
                    Arrays.copyOfRange(data, nAlleles*j, nAlleles*(j+1))));
        }
        return sb.toString();
    }
}
