/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import beagleutil.Samples;
import java.io.*;
import java.util.*;
import blbutil.*;

/**
 * Class <code>SignalsIterator</code> iterates through data in a file
 * in BEAGLE .signals format and returns a <code>SignalsData</code> instance
 * for each marker in the .signals file.
 *
 * @author Brian L Browning
 */
public class SignalsIterator implements FileIterator<SignalsData> {

    private final Samples sampleIds;
    private final boolean isMultiAllelic;
    private final String signalsFilename;
    private final FileIterator<String> it;
    private final ImmutableSet<String> markerExclusions;

    private SignalsData next;

    /**
     * Constructs a <code>SignalsIterator</code> instance.
     *
     * @param signalsFile a file in BEAGLE .signals format.  If
     * <code>isMultiAllelic==true</code> then the first (header) line of the
     * specified .gprobs file consists of a white-space delimited list
     * of ordered sample identifiers.  If <code>isMultiAllelic==false</code>
     * then the first (header) line begins with "marker alleleA alleleB "
     * and continues with the sample identifiers corresponding to each
     * column of input data (i.e. each sample identifier is repeated two
     * times).
     * @param sampleExclusions sample identifiers to be excluded
     * @param markerExclusions marker identifiers to be excluded
     * @param isMultiAllelic returns <code>true</code> if the input data
     * files contain a field specifying the number of alleles and
     * <code>false</code> otherwise.  If <code>isMultiAllelic==false</code>
     * the marker is assumed to be diallelic.
     *
     * @throws NullPointerException if
     * <code>signalsFiles==null || sampleExclusions==null
     * || markerExclusions==null</code>.
     *
     * @throws IllegalArgumentException if the .gprobs file is not
     * in the correct format.
     *
     */
    public SignalsIterator(File signalsFile,
            ImmutableSet<String> sampleExclusions,
            ImmutableSet<String> markerExclusions, boolean isMultiAllelic) {
        if (markerExclusions==null) {
            throw new NullPointerException("markerExclusions==null");
        }
        this.isMultiAllelic = isMultiAllelic;
        this.markerExclusions = markerExclusions;
        this.signalsFilename = signalsFile.getAbsolutePath();
        this.it = InputStreamIterator.fromGzipFile(signalsFile);
        String headerLine = readFirstNonBlankLine(this.it);
        String[] ids = sampleIds(headerLine, isMultiAllelic);
        this.sampleIds = new Samples(ids, sampleExclusions);
        this.next = readData();
    }

    private String readFirstNonBlankLine(FileIterator<String> it) {
        String line = null;
        while (it.hasNext() && line==null) {
            String candidate = it.next().trim();
            if (candidate.length() > 0) {
                line = candidate;
            }
        }
        if (line==null) {
            String s = " missing data in file: " + signalsFilename;
            throw new IllegalArgumentException(s);
        }
        return line;
    }

    private String[] sampleIds(String headerLine, boolean isMultiAllelic) {
        String[] headerFields = StringUtil.getFields(headerLine);
        if (isMultiAllelic) {
            return headerFields;
        }
        else {
            int idIndex = 0;
            int startIndex = 3;
            if (((headerFields.length - 3) % 2) != 0) {
                String s = "number of fields in header line of "
                        + signalsFilename + " is not odd: "
                        + headerFields;
                throw new IllegalArgumentException(s);
            }
            String[] ids = new String[(headerFields.length - 3) / 2];
            for (int j=startIndex; j<headerFields.length; j+=2) {
                if (!headerFields[j].equals(headerFields[j+1])) {
                    String s = "Require sets of 2 identical sample " +
                            "identifiers in " + signalsFilename +
                            " header line: " + headerFields[j]
                            + " != " + headerFields[j+1];
                    throw new IllegalArgumentException(s);
                }
                ids[idIndex++] = headerFields[j];
            }
            assert idIndex == ids.length;
            return ids;
        }
    }

    public void close() {
        it.close();
        next = null;
    }

    public boolean hasNext() {
        return (next != null);
    }

    public SignalsData next() {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        SignalsData current = next;
        next = readData();
        return current;
    }

    public void remove() {
        String s = "remove() is not supported by SignalsIterator";
        throw new UnsupportedOperationException(s);
    }

    private SignalsData readData() {
        SignalsData data = null;
        while (it.hasNext() && data==null) {
            String candidate = it.next();
            String[] fields = StringUtil.getFields(candidate, 2);
            if (fields.length > 0 && !markerExclusions.contains(fields[0])) {
                data = new SignalsData(candidate, sampleIds, isMultiAllelic);
            }
        }
        return data;
    }

    /**
     * Returns a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to change.
     * @return a string representation of <code>this</code>.
     */
    @Override
    public String toString() {
        String space = " ";
        StringBuilder sb = new StringBuilder();
        sb.append(".signals files: ");
        sb.append(signalsFilename);
        sb.append(space);
        sb.append("isMultialllelic=");
        sb.append(isMultiAllelic);
        sb.append(space);
        sb.append(" next=");
        sb.append(next);
        return sb.toString();
    }
}
