/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import java.util.Arrays;

import org.apache.commons.math.linear.*;
import org.apache.commons.math.special.*;

/**
 * Probability density for a multivariate t distribution.
 * 
 * @author Brian L Browning
 */
public class T implements ProbabilityDensity {

    private final double df;
    private final double[] mean;
    private final double c;
    private final RealMatrix variance;
    private final RealMatrix varInverse;

    /**
     * Constructs a <code>ProbabilityDensity</code> object
     * for a multivariate t-distribution.
     * 
     * @param mean the distribution mean.
     * @param variance the matrix that parameterizes the 
     * variance/covariance matrix for the distribution.  The variance of the
     * t-distribution is obtained by multiplying each element of
     * the specified <code>variance</code> parameters by
     * <code>(df/(df - 2.0)</code>.
     * @param df the degrees of freedom.
     *
     * @throws IllegalArgumentException if
     * <code>mean.length != variance.length</code> or if any element of the
     * mean and variance arrays are not finite numbers or if
     * <code>df <= 0.0</code>.
     * @throws NullPointerException if
     * <code>mean==null || variance==null || variance[j]==null</code>
     * for any <code>0 &le; j < variance.length</code>.
     */
    public T(double[] mean, double[][] variance, double df) {
        if (df <= 0.0) {
            throw new IllegalArgumentException("df <= 0.0: " + df);
        }
        checkDimensions(mean.length, variance.length);
        for (double[] da : variance) {
            checkDimensions(mean.length, da.length);
        }
        checkEntries(mean);
        for (double[] da : variance) {
            checkEntries(da);
        }
        this.df = df;
        this.mean = mean.clone();
        this.variance = new RealMatrixImpl(variance);
        this.varInverse = this.variance.inverse();

        double d = ratioOfGammas(df, mean.length);
        d /= Math.pow(this.variance.getDeterminant(), 0.5);
        d /= Math.pow((Math.PI * df), mean.length/2.0);
        this.c = d;
    }

    private void checkDimensions(int d1, int d2) {
        if (d1!=d2) {
            String s = "dimensions are inconsistent: " + d1 + " " + d2;
            throw new IllegalArgumentException(s);
        }
    }

    private void checkEntries(double[] da) {
        for (double d : da) {
            if (Double.isNaN(d)) {
                String s = "ERROR: undefined parameter: " + Arrays.toString(da);
                throw new IllegalArgumentException(s);
            }
            if (Double.isInfinite(d)) {
                String s = "Error: infinite parameter: " + Arrays.toString(da);
                throw new IllegalArgumentException(s);
            }
        }
    }

    /*
     * Returns Gamma[df/2.0] / Gamma[(df/2.0) + (dimension/2.0)]
     */
    private static double ratioOfGammas(double df, int dimension) {
        assert dimension >= 0;
        double value = 1.0;
        for (int j=(dimension - 2); j>=0; j-=2) {
            value *= (0.5)*(df + j);
        }
        if ((dimension & 1) != 0) {
            while (df > 4.0) {
                df -= 2.0;
                value *= ((df + 1.0)/df);
            }
            value *= Math.exp(Gamma.logGamma((df + 1.0)/2.0) - Gamma.logGamma(df/2.0));
        }
        return value;
    }

    public double probability(double[] x) {
        if (x.length != mean.length) {
            String s = "x.length != dimension(): " + x.length + " " + dimension();
            throw new IllegalArgumentException(s);
        }
        double quadForm = 0;
        double[] shiftedX = new double[x.length];
        for (int j=0; j<mean.length; ++j) {
            if (Double.isNaN(x[j])) {
                return Double.NaN;
            }
            if (Double.isInfinite(x[j])) {
                return 0.0;
            }
            shiftedX[j] = x[j] - mean[j];
        }
        for (int i=0; i<mean.length; ++i) {
            for (int j=0; j<mean.length; ++j) {
                quadForm += shiftedX[i]*varInverse.getEntry(i, j)*shiftedX[j];
            }
        }
        return (c*Math.pow((1 + quadForm/df), -(0.5)*(df + mean.length)));
    }

    public int dimension() {
        return mean.length;
    }

    public double[] mean() {
        return mean.clone();
    }

    public double[][] variance() {
        return variance.getData();
    }

    public double df() {
        return df;
    }

    public String name() {
        return "t-dist";
    }

    /**
     * Returns a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to
     * change.
     *
     * @return a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to
     * change.
     */
    @Override
    public String toString() {
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        sb.append("T Density");
        sb.append(newLine);
        sb.append("df=" + df);
        sb.append(newLine);
        sb.append("mean=" + Arrays.toString(mean));
        sb.append(newLine);
        sb.append("variance=" + variance);
        sb.append(newLine);
        sb.append("varInverse=" + varInverse);
        sb.append(newLine);
        return sb.toString();
    }

// <editor-fold defaultstate="collapsed" desc=" commented-out code for testing class ">
/* code for testing class */
//    public static void main(String[] args) {
//        double df = 2000;
//        int dim = 2;
//        double a = Gamma.gamma((0.5)*(df + dim)) / Gamma.gamma(0.5 * df);
//        double b = ratioOfGammas(df, dim);
//        System.out.println("a=" + a + "  b=" + b + "  : (a-b)/b = " + ((a-b)/b));
//
//    }
// </editor-fold>
}
