/*
 * Copyright 2009 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package calling;

import java.util.Arrays;

/**
 * Probability density for a multivariate uniform distribution with
 * independent marginal distributions.  Points that lie on the
 * boundary of the support region are assigned nonzero probability.
 * 
 * @author Brian L Browning
 */
public final class Uniform implements ProbabilityDensity {

    private final double[][] support;
    private final double prob;

    /**
     * Constructs a <code>ProbabilityDensity</code> object
     * for a multivariate uniform distribution with independent
     * marginal distributions.
     *
     * @param support a two-dimensional array with two columns such 
     * that <code>support[j][0]</code> is the minimum support for the
     * <code>j</code>-th dimension and <code>support[j][1]</code> is
     * the maximum support for the <code>j</code>-th dimension for
     * <code>0 &le j < support.length</code>.  The
     * number of rows of the array equals the number of dimensions.
     *
     * @throws IllegalArgumentException if
     * <code>support.length == 0</code> or if
     * <code>support[j].length != 2</code> for any
     * <code>0 &le; j < support.length</code> or if 
     * any element of the specified arrays is not a finite number, or
     * if <code>support[0][j] >= support[1][j]</code> for any element
     * <code>0 &le; j < support.length</code>.
     * @throws NullPointerException if <code>support==null</code>
     * or if <code>support[j]==null</code> for 
     * <code>0 &le; j < support.length</code>.
     */
    public Uniform(double[][] support) {
        if (support.length == 0) {
            throw new IllegalArgumentException("support.length==0");
        }
        for (double[] da : support) {
            checkFinite(da);
            checkRange(da);
        }
        this.support = new double[support.length][];
        for (int j=0; j<support.length; ++j) {
            this.support[j] = support[j].clone();
        }
        double volume = 1.0;
        for (int j=0; j<support.length; ++j) {
            volume *= (support[j][1] - support[j][0]);
        }
        this.prob = 1.0/volume;
    }

    private void checkFinite(double[] da) {
        for (double d : da) {
            if (Double.isNaN(d) || Double.isInfinite(d)) {
                String s = "array element must be finite: "
                        + java.util.Arrays.toString(da);
                throw new IllegalArgumentException(s);
            }
        }
    }

    private void checkRange(double[] da) {
        if (da.length != 2) {
            String s = "expected two-element array: " + Arrays.toString(da);
            throw new IllegalArgumentException(s);
        }
        if (da[0] >= da[1]) {
            String s = "support array is not increasing: " + Arrays.toString(da);
            throw new IllegalArgumentException(s);
        }
    }
    
    public double probability(double[] x) {
        if (x.length != dimension()) {
            String s = "x.length=" + x.length + " != dimension()=" + dimension();
            throw new IllegalArgumentException(s);
        }
        for (int j=0; j<x.length; ++j) {
            if (x[j] < support[j][0] || x[j] > support[j][1]) {
                return 0.0;
            }
        }
        return prob;
    }

    public int dimension() {
        return support.length;
    }

    public double[] mean() {
        double[] mean = new double[dimension()];
        for (int j=0; j<mean.length; ++j) {
            mean[j] = (support[j][1] + support[j][0])/2.0;
        }
        return mean;
    }

    public double[][] variance() {
        double[][] var = new double[dimension()][dimension()];
        for (int j=0; j<var.length; ++j) {
            double x = (support[j][1] - support[j][0]);
            var[j][j] = x*x/12.0;
        }
        return var;
    }

    public double df() {
        return Double.NaN;
    }

    public String name() {
        return "uniform";
    }

    /**
     * Returns a string representation of <code>this</code>.  The exact
     * details of the representation are unspecified and subject to change.
     * @return a string representation of <code>this</code>.
     */
    @Override
    public String toString() {
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        for (int j=0; j<support.length; ++j) {
            sb.append("dim ");
            sb.append(j);
            sb.append(": ");
            sb.append(Arrays.toString(support[j]));
            sb.append(newLine);
        }
        sb.append("value: ");
        sb.append(prob);
        return sb.toString();
    }
}
