 package beagleutil;

/**
 * The algorithm for computing P-values from an exact HWE test is given in
 * the article: JE Wigginton, DJ Cutler, GR Abecasis (2005) A Note on
 * exact tests of Hardy-Weinberg equilibrium.  Am J Hum Genet 76:887-893.
 *
 * The code in SNP_HWE.hweExact() is adapted from code available from
 * G. Abecasis's web site (see URL in referenced article), and is made
 * available with the following permissions:
 * 
 * "With appropriate citation, these routines are freely available for 
 * your use and can be incorporated into other programs."
 * Source: http://www.sph.umich.edu/csg/abecasis/Exact/ (accessed 19 Nov 2009).
 *
 * @author Jan Wigginton wrote SNP-HWE in C++, and the C++ verision was
 * ported to java by Brian L Browning.
 *
 */
public class SNP_HWE {
    /**
     * Returns the p-value from an exact test of Hardy-Weinberg Disequilibrium.
     * @param obsAA homozygote count for the reference allele
     * @param obsAB heterozygote count
     * @param obsBB homozygote count for the non-reference allele.
     * @return the p-value from an exact test of Hardy-Weinberg Disequilibrium.
     * @throws IllegalArgumentException if
     * <code>count0 < 0 || count1 < 0 || count2 < 0</code>.
     *
     * Reference: JE Wigginton, DJ Cutler, GR Abecasis (2005) A Note on
     * exact tests of Hardy-Weinberg equilibrium.  Am J Hum Genet 76:887-893.
     *
     * The following code is adapted from code available from G Abecasis'
     * web site (see URL in referenced article).
     */
    public static double hweExact(int obsAA, int obsAB, int obsBB) {
        if (obsAA < 0 || obsAB < 0 || obsBB < 0) {
            String s = "counts cannot be negative: " + obsAA + " " + obsAB
                    + " " + obsBB;
            throw new IllegalArgumentException(s);
        }
        int n = obsAA + obsAB + obsBB;
        if (n==0) {
            return 1.0;
        }
        int obsA = 2*obsAA + obsAB;
        int obsB = 2*obsBB + obsAB;
        double[] probAB = new double[Math.min(obsA, obsB) + 1];

        int expAB = (obsA*obsB)/(2*n);  // expected AB count
        if (((obsB & 1)==1) ^ ((expAB & 1)==1)) {
            ++expAB; // expAB and obsB should both be even or both be odd
        }
        int expAA = (obsB - expAB)/2;
        int expBB = n - expAB - expAA;

        probAB[expAB] = 1.0;
        double sum = probAB[expAB];
        for (int nAA=expAA, nAB=expAB, nBB=expBB; nAB > 1; nAB -= 2) {
            probAB[nAB - 2] = probAB[nAB] * nAB * (nAB - 1.0) /
                    (4.0 * (nAA + 1.0) * (nBB + 1.0));
            sum += probAB[nAB - 2];
            nAA++;
            nBB++;
        }
        for (int nAA=expAA, nAB=expAB, nBB=expBB, max=probAB.length - 3;
                nAB <= max; nAB += 2) {
            probAB[nAB + 2] = probAB[nAB] * 4.0 * nAA * nBB
                            /((nAB + 2.0) * (nAB + 1.0));
            sum += probAB[nAB + 2];
            nAA--;
            nBB--;
        }

        double hweP = 0.0;
        for (int j=0; j<probAB.length; ++j) {
            if (probAB[j] <= probAB[obsAB]) {
                hweP += probAB[j];
            }
        }
        hweP /= sum;
        return (hweP > 1.0) ? 1.0 : hweP;
    }
}
